from typing import Dict, Any
from textual.widgets import Static, DataTable
from textual.app import App, ComposeResult


class PositionsWidget(Static):
    """Widget for displaying positions"""

    def __init__(self):
        super().__init__()
        self.add_class("widget")

    def compose(self):
        table = DataTable(classes="positions-table")
        table.add_columns("Symbol", "Side", "Amount", "Entry Price")
        yield table

    def update_data(self, data: Dict[str, Any]):
        """Update positions table with new data"""
        table = self.query_one(DataTable)
        table.clear()

        positions = data.get("positions", [])

        for position in positions:
            symbol = position.get("symbol", "N/A")
            side = position.get("side", "N/A")
            amount = (
                f"{position.get('amount', 0):.6f}" if position.get("amount") else "N/A"
            )
            entry_price = (
                f"{position.get('entry_price', 0):.6f}"
                if position.get("entry_price")
                else "N/A"
            )

            table.add_row(symbol, side, amount, entry_price)


class PositionsTestApp(App):
    """Test application for PositionsWidget"""

    CSS = """
    Screen {
        background: $surface;
    }
    
    .widget {
        border: solid $primary;
        height: auto;
        margin: 1 2;
    }
    
    .positions-table {
        height: auto;
    }
    """

    def compose(self) -> ComposeResult:
        from textual.widgets import Header, Footer

        yield Header()
        yield PositionsWidget()
        yield Footer()

    def on_mount(self):
        self.set_interval(2, self.load_sample_data)

    def load_sample_data(self):
        """Load sample positions data into the widget"""
        from random import random, choice

        sample_data = {
            "positions": [
                {
                    "symbol": "BTCUSDT-PERP",
                    "side": choice(["LONG", "SHORT"]),
                    "amount": 1.5 + random() * 3,
                    "entry_price": 95000.0 + random() * 1000,
                    "mark_price": 95500.0 + random() * 1000,
                    "unrealized_pnl": (random() - 0.5) * 2000,
                    "roe_percent": (random() - 0.5) * 20,
                },
                {
                    "symbol": "ETHUSDT-PERP",
                    "side": choice(["LONG", "SHORT"]),
                    "amount": 10.0 + random() * 15,
                    "entry_price": 3500.0 + random() * 200,
                    "mark_price": 3520.0 + random() * 200,
                    "unrealized_pnl": (random() - 0.5) * 1500,
                    "roe_percent": (random() - 0.5) * 15,
                },
                {
                    "symbol": "ADAUSDT-PERP",
                    "side": choice(["LONG", "SHORT"]),
                    "amount": 5000.0 + random() * 3000,
                    "entry_price": 1.2 + random() * 0.3,
                    "mark_price": 1.25 + random() * 0.3,
                    "unrealized_pnl": (random() - 0.5) * 500,
                    "roe_percent": (random() - 0.5) * 25,
                },
            ]
        }

        positions_widget = self.query_one(PositionsWidget)
        positions_widget.update_data(sample_data)


if __name__ == "__main__":
    app = PositionsTestApp()
    app.run()



import asyncio
import nexuslog as logging

from typing import Callable, Awaitable, Generic, TypeVar
from random import randint

T = TypeVar("T")


def get_exponential_backoff(
    num_attempts: int,
    delay_initial_ms: int = 500,
    delay_max_ms: int = 2_000,
    backoff_factor: int = 2,
    jitter: bool = True,
) -> int:
    """
    Compute the backoff using exponential backoff and jitter.

    Parameters
    ----------
    num_attempts : int, default 1
        The number of attempts that have already been made.
    delay_initial_ms : int, default 500
        The time to sleep in the first attempt.
    delay_max_ms : int, default 2_000
        The maximum delay.
    backoff_factor : int, default 2
        The exponential backoff factor for delays.
    jitter : bool, default True
        Whether or not to apply jitter.

    Notes
    -----
    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/

    Returns
    -------
    int
        Delay in milliseconds.

    """
    delay = min(delay_max_ms, delay_initial_ms * backoff_factor ** (num_attempts - 1))

    if jitter:
        return randint(delay_initial_ms, delay)  # noqa: S311

    return delay


class RetryManager(Generic[T]):
    """
    Provides retry state management for an HTTP request.

    This class is generic over `T`, where `T` is the return type of the
    function passed to the `run` method.

    Parameters
    ----------
    max_retries : int
        The maximum number of retries before failure.
    delay_initial_ms : int
        The initial delay (milliseconds) for retries.
    delay_max_ms : int
        The maximum delay (milliseconds) for exponential backoff.
    backoff_factor : int
        The exponential backoff factor for retry delays.
    exc_types : tuple[Type[BaseException], ...]
        The exception types to handle for retries.
    retry_check : Callable[[BaseException], None], optional
        A function that performs additional checks on the exception.
        If the function returns `False`, a retry will not be attempted.

    """

    def __init__(
        self,
        max_retries: int,
        delay_initial_ms: int,
        delay_max_ms: int,
        backoff_factor: int,
        exc_types: tuple[type[BaseException], ...],
        retry_check: Callable[[BaseException], bool] | None = None,
    ) -> None:
        self.max_retries = max_retries
        self.delay_initial_ms = delay_initial_ms
        self.delay_max_ms = delay_max_ms
        self.backoff_factor = backoff_factor
        self.exc_types = exc_types
        self.retry_check = retry_check
        self.cancel_event = asyncio.Event()
        self._log = logging.getLogger(name=type(self).__name__)

        self.name: str | None = None
        self.details: list[object] | None = None
        self.details_str: str | None = None
        self.result: bool = False
        self.message: str | None = None

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(name='{self.name}', details={self.details}) at {hex(id(self))}>"

    async def run(
        self,
        name: str,
        func: Callable[..., Awaitable[T]],
        details: list[object] | None = None,
        jitter: bool = True,
        *args,
        **kwargs,
    ) -> T | None:
        """
        Execute the given `func` with retry management.

        If an exception in `self.exc_types` is raised, a warning is logged, and the function is
        retried after a delay until the maximum retries are reached, at which point an error is logged.

        Parameters
        ----------
        name : str
            The name of the operation to run.
        details : list[object], optional
            The operation details such as identifiers.
        func : Callable[..., Awaitable[T]]
            The function to execute.
        jitter : bool, default True
            Whether to apply jitter to the retry delay.
        args : Any
            Positional arguments to pass to the function `func`.
        kwargs : Any
            Keyword arguments to pass to the function `func`.

        Returns
        -------
        T | None
            The result of the executed function, or ``None`` if the retries fail.

        """
        retries = 0
        self.name = name
        self.details = details

        try:
            while True:
                if self.cancel_event.is_set():
                    self._cancel()
                    return None

                try:
                    response = await func(*args, **kwargs)
                    self.result = True

                    return response  # Successful request
                except self.exc_types as e:
                    self._log.warning(repr(e))
                    if (
                        (self.retry_check and not self.retry_check(e))
                        or not self.max_retries
                        or retries >= self.max_retries
                    ):
                        self._log_error()
                        self.result = False
                        self.message = str(e)
                        raise e

                    retries += 1
                    retry_delay_ms = get_exponential_backoff(
                        delay_initial_ms=self.delay_initial_ms,
                        delay_max_ms=self.delay_max_ms,
                        backoff_factor=self.backoff_factor,
                        num_attempts=retries,
                        jitter=jitter,
                    )
                    self._log_retry(retries, retry_delay_ms=retry_delay_ms)
                    await asyncio.sleep(retry_delay_ms / 1000)
        except asyncio.CancelledError:
            self._cancel()
            return None

    def cancel(self) -> None:
        """
        Cancel the retry operation.
        """
        self._log.debug(f"Canceling {self!r}")
        self.cancel_event.set()

    def clear(self) -> None:
        """
        Clear all state from this retry manager.
        """
        self.name = None
        self.details = None
        self.details_str = None
        self.result = False
        self.message = None

    def _cancel(self) -> None:
        self._log.warning(f"Canceled retry for '{self.name}'")
        self.result = False
        self.message = "Canceled retry"

    def _log_retry(self, retry: int, retry_delay_ms: int) -> None:
        self._log.warning(
            f"Retrying {retry}/{self.max_retries} for '{self.name}' "
            f"in {retry_delay_ms / 1000}s{self._details_str()}",
        )

    def _log_error(self) -> None:
        self._log.error(
            f"Failed on '{self.name}'{self._details_str()}",
        )

    def _details_str(self) -> str:
        if not self.details:
            return ""

        if not self.details_str:
            self.details_str = ": " + ", ".join([repr(x) for x in self.details])

        return self.details_str



import pandas as pd
from decimal import Decimal
from typing import Dict, List, Any
from typing import Optional
from msgspec import Struct, field
from nexustrader.core.nautilius_core import UUID4
from nexustrader.constants import (
    OrderSide,
    OrderType,
    TimeInForce,
    OrderStatus,
    PositionSide,
    InstrumentType,
    ExchangeType,
    AlgoOrderStatus,
    KlineInterval,
    TriggerType,
    DataType,
)


class BatchOrder(Struct, kw_only=True):
    symbol: str
    side: OrderSide
    type: OrderType
    amount: Decimal
    price: Decimal | None = None  # for market order, the price is None
    time_in_force: TimeInForce | None = TimeInForce.GTC
    reduce_only: bool = False
    kwargs: Dict[str, Any] = field(default_factory=dict)


class Symbol(str):
    """Symbol class that inherits from string with all InstrumentId methods."""

    def __new__(cls, value: str):
        instance = str.__new__(cls, value)
        instance._instrument_id = None
        return instance

    def __init__(self, value: str):
        super().__init__()
        self._instrument_id = InstrumentId.from_str(value)

    @property
    def is_spot(self) -> bool:
        return self._instrument_id.is_spot

    @property
    def is_linear(self) -> bool:
        return self._instrument_id.is_linear

    @property
    def is_inverse(self) -> bool:
        return self._instrument_id.is_inverse

    @property
    def id(self) -> str:
        """Get the id from the symbol."""
        return self._instrument_id.id

    @property
    def exchange(self) -> ExchangeType:
        """Get the exchange from the symbol."""
        return self._instrument_id.exchange

    @property
    def type(self) -> InstrumentType:
        """Get the instrument type from the symbol."""
        return self._instrument_id.type


class InstrumentId(Struct):
    id: str
    symbol: str
    exchange: ExchangeType
    type: InstrumentType

    @property
    def is_spot(self) -> bool:
        return self.type == InstrumentType.SPOT

    @property
    def is_linear(self) -> bool:
        return self.type == InstrumentType.LINEAR

    @property
    def is_inverse(self) -> bool:
        return self.type == InstrumentType.INVERSE

    @classmethod
    def from_str(cls, symbol: str):
        """
        BTCETH.BINANCE -> SPOT
        BTCUSDT-PERP.BINANCE -> LINEAR
        BTCUSD.BINANCE -> INVERSE
        BTCUSD-241227.BINANCE
        """
        symbol_prefix, exchange = symbol.split(".")

        # if numirical number in id, then it is a future
        if "-" in symbol_prefix:
            prefix, _ = symbol_prefix.split("-")
            if prefix.endswith("USD"):
                type = InstrumentType.INVERSE
            else:
                type = InstrumentType.LINEAR
        else:
            type = InstrumentType.SPOT

        return cls(
            id=symbol_prefix,
            symbol=symbol,
            exchange=ExchangeType(exchange.lower()),
            type=type,
        )


class BookL1(Struct, gc=False, frozen=True):
    exchange: ExchangeType
    symbol: str
    bid: float
    ask: float
    bid_size: float
    ask_size: float
    timestamp: int

    @property
    def mid(self) -> float:
        return (self.bid + self.ask) / 2

    @property
    def spread(self) -> float:
        return self.ask - self.bid

    @property
    def weighted_mid(self) -> float:
        return (self.bid * self.ask_size + self.ask * self.bid_size) / (
            self.ask_size + self.bid_size
        )


class BookOrderData(Struct, gc=False, frozen=True):
    price: float
    size: float


class BookL2(Struct, frozen=True):
    exchange: ExchangeType
    symbol: str
    bids: List[BookOrderData]  # desc order
    asks: List[BookOrderData]  # asc order
    timestamp: int


class Trade(Struct, gc=False, frozen=True):
    exchange: ExchangeType
    side: OrderSide
    symbol: str
    price: float
    size: float
    timestamp: int


class Kline(Struct, gc=False, kw_only=True, frozen=True, omit_defaults=True):
    exchange: ExchangeType
    symbol: str
    interval: KlineInterval
    open: float
    high: float
    low: float
    close: float
    volume: float | None = None
    buy_volume: float | None = None  # for trade aggregation only add trade.side == BUY
    quote_volume: float | None = None  # only for binance and okx
    taker_volume: float | None = None  # only for binance
    taker_quote_volume: float | None = None  # only for binance
    turnover: float | None = None  # only for bybit
    start: int
    timestamp: int
    confirm: bool


class MarkPrice(Struct, gc=False, frozen=True):
    exchange: ExchangeType
    symbol: str
    price: float
    timestamp: int


class FundingRate(Struct, gc=False, frozen=True):
    exchange: ExchangeType
    symbol: str
    rate: float
    timestamp: int
    next_funding_time: int


class Ticker(Struct, gc=False, kw_only=True):
    """
    Universal ticker data structure supporting all exchanges.
    volumeCcy: for linear and spot markets, it is the volume in quote currency.
    For inverse markets, it is the volume in base currency.
    """

    exchange: ExchangeType
    symbol: str
    last_price: float
    timestamp: int
    volume: float
    volumeCcy: float


class IndexPrice(Struct, gc=False, frozen=True):
    exchange: ExchangeType
    symbol: str
    price: float
    timestamp: int


class OrderSubmit(Struct):
    symbol: str
    instrument_id: InstrumentId
    kwargs: Dict[str, Any] = field(default_factory=dict)


class BatchOrderSubmit(OrderSubmit, kw_only=True):
    oid: str
    side: OrderSide
    type: OrderType
    amount: Decimal
    price: Decimal | None = None  # for market order, the price is None
    time_in_force: TimeInForce = TimeInForce.GTC
    reduce_only: bool = False


class CreateOrderSubmit(OrderSubmit, kw_only=True):
    oid: str
    side: OrderSide
    type: OrderType
    amount: Decimal
    price: Decimal | None = None  # for market order, the price is None
    time_in_force: TimeInForce | None = TimeInForce.GTC
    reduce_only: bool = False
    # position_side: PositionSide | None = None


class CancelOrderSubmit(OrderSubmit, kw_only=True):
    oid: str


class CancelAllOrderSubmit(OrderSubmit, kw_only=True):
    pass


class TakeProfitAndStopLossOrderSubmit(CreateOrderSubmit, kw_only=True):
    tp_order_type: OrderType
    tp_trigger_price: Decimal | None = None
    tp_price: Decimal | None = None
    tp_trigger_type: TriggerType = TriggerType.LAST_PRICE
    sl_order_type: OrderType
    sl_trigger_price: Decimal | None = None
    sl_price: Decimal | None = None
    sl_trigger_type: TriggerType = TriggerType.LAST_PRICE


class TWAPOrderSubmit(OrderSubmit, kw_only=True):
    uuid: str = field(default_factory=lambda: f"ALGO-{UUID4().value}")
    side: OrderSide
    amount: Decimal
    duration: int
    wait: int
    position_side: PositionSide
    check_interval: float = 0.1


class CancelTWAPOrderSubmit(OrderSubmit, kw_only=True):
    uuid: str


class ModifyOrderSubmit(OrderSubmit, kw_only=True):
    oid: str
    side: OrderSide
    price: Decimal
    amount: Decimal


class SubscriptionSubmit(Struct):
    symbols: List[str]
    data_type: DataType
    params: Dict[str, Any] = field(default_factory=dict)
    ready_timeout: int = 60
    ready: bool = True


class UnsubscriptionSubmit(Struct):
    symbols: List[str]
    data_type: DataType
    params: Dict[str, Any] = field(default_factory=dict)


class Order(Struct):
    exchange: ExchangeType
    symbol: str
    status: OrderStatus
    oid: Optional[str] = None
    eid: Optional[str] = None
    amount: Optional[Decimal] = None
    filled: Optional[Decimal] = None
    timestamp: Optional[int] = None
    type: Optional[OrderType] = None
    side: Optional[OrderSide] = None
    time_in_force: Optional[TimeInForce] = None
    price: Optional[float] = None
    trigger_price: Optional[float] = None
    average: Optional[float] = None
    last_filled_price: Optional[float] = None
    last_filled: Optional[Decimal] = None
    remaining: Optional[Decimal] = None
    fee: Optional[Decimal] = None
    fee_currency: Optional[str] = None
    cost: Optional[Decimal] = None
    cum_cost: Optional[Decimal] = None
    reduce_only: Optional[bool] = None
    position_side: Optional[PositionSide] = None

    @property
    def success(self) -> bool:
        return self.status not in [OrderStatus.FAILED, OrderStatus.CANCEL_FAILED]

    @property
    def is_filled(self) -> bool:
        return self.status == OrderStatus.FILLED

    @property
    def is_pending(self) -> bool:
        return self.status == OrderStatus.PENDING

    @property
    def is_accepted(self) -> bool:
        return self.status == OrderStatus.ACCEPTED

    @property
    def is_partially_filled(self) -> bool:
        return self.status == OrderStatus.PARTIALLY_FILLED

    @property
    def is_partially_canceled(self) -> bool:
        return self.status == OrderStatus.CANCELED and self.filled > Decimal("0")

    @property
    def is_canceled(self) -> bool:
        return self.status == OrderStatus.CANCELED

    @property
    def is_closed(self) -> bool:
        return self.status in [
            OrderStatus.FILLED,
            OrderStatus.CANCELED,
            OrderStatus.EXPIRED,
            OrderStatus.CANCEL_FAILED,
            OrderStatus.FAILED,
        ]

    @property
    def is_opened(self) -> bool:
        return self.status in [
            OrderStatus.PENDING,
            OrderStatus.CANCELING,
            OrderStatus.PARTIALLY_FILLED,
            OrderStatus.ACCEPTED,
        ]

    @property
    def on_flight(self) -> bool:
        return self.status in [
            OrderStatus.PENDING,
            OrderStatus.CANCELING,
        ]

    @property
    def is_buy(self) -> bool:
        return self.side == OrderSide.BUY

    @property
    def is_sell(self) -> bool:
        return self.side == OrderSide.SELL

    @property
    def is_maker(self) -> bool:
        return self.type == OrderType.LIMIT

    @property
    def is_taker(self) -> bool:
        return self.type == OrderType.MARKET

    @property
    def is_post_only(self) -> bool:
        return self.type == OrderType.POST_ONLY


class AlgoOrder(Struct, kw_only=True):
    symbol: str
    oid: str  # start with "ALGO-"
    side: OrderSide
    amount: Optional[Decimal] = None
    duration: int
    wait: int
    status: AlgoOrderStatus
    exchange: ExchangeType
    timestamp: int
    orders: List[str] = field(default_factory=list)  # [uuid1, uuid2, ...]
    position_side: PositionSide | None = None
    filled: Optional[Decimal] = None
    cost: Optional[float] = None
    average: Optional[float] = None

    @property
    def success(self) -> bool:
        return not self.is_failed

    @property
    def is_running(self) -> bool:
        return self.status == AlgoOrderStatus.RUNNING

    @property
    def is_finished(self) -> bool:
        return self.status == AlgoOrderStatus.FINISHED

    @property
    def is_canceled(self) -> bool:
        return self.status == AlgoOrderStatus.CANCELED

    @property
    def is_failed(self) -> bool:
        return self.status == AlgoOrderStatus.FAILED

    @property
    def is_closed(self) -> bool:
        return self.status in [
            AlgoOrderStatus.CANCELED,
            AlgoOrderStatus.FAILED,
            AlgoOrderStatus.FINISHED,
        ]

    @property
    def is_opened(self) -> bool:
        return self.status in [AlgoOrderStatus.RUNNING, AlgoOrderStatus.CANCELING]


class Balance(Struct):
    """
    Buy BTC/USDT: amount = 0.01, cost: 600

    OrderStatus.INITIALIZED: BTC(free: 0.0, locked: 0.0) USDT(free: 1000, locked: 0)
    OrderStatus.PENDING: BTC(free: 0.0, locked: 0) USDT(free: 400, locked: 600) USDT.update_locked(600) USDT.update_free(-600)

    OrderStatus.PARTIALLY_FILLED: BTC(free: 0.005, locked: 0) USDT(free: 400, locked: 300) BTC.update_free(0.005) USDT.update_locked(-300)
    OrderStatus.FILLED: BTC(free: 0.01, locked: 0.0) USDT(free: 400, locked: 0) BTC.update_free(0.005) USDT.update_locked(-300)

    Buy BTC/USDT: amount = 0.01, cost: 200

    OrderStatus.INITIALIZED: BTC(free: 0.01, locked: 0.0) USDT(free: 400, locked: 0)
    OrderStatus.PENDING: BTC(free: 0.01, locked: 0.0) USDT(free: 200, locked: 200) USDT.update_locked(200) USDT.update_free(-200)
    OrderStatus.FILLED: BTC(free: 0.02, locked: 0.0) USDT(free: 200, locked: 0) BTC.update_free(0.01) USDT.update_locked(-200)

    Sell BTC/USDT: amount = 0.01, cost: 300
    OrderStatus.INITIALIZED: BTC(free: 0.02, locked: 0.0) USDT(free: 200, locked: 0)
    OrderStatus.PENDING: BTC(free: 0.01, locked: 0.01) USDT(free: 200, locked: 0) BTC.update_locked(0.01) BTC.update_free(-0.01)
    OrderStatus.PARTIALLY_FILLED: BTC(free: 0.01, locked: 0.005) USDT(free: 350, locked: 0) BTC.update_locked(-0.005) USDT.update_free(150)
    OrderStatus.FILLED: BTC(free: 0.01, locked: 0.0) USDT(free: 500, locked: 0) BTC.update_locked(-0.005) USDT.update_free(150)
    """

    asset: str
    free: Decimal = field(default=Decimal("0.0"))
    locked: Decimal = field(default=Decimal("0.0"))

    @property
    def total(self) -> Decimal:
        return self.free + self.locked


class AccountBalance(Struct):
    balances: Dict[str, Balance] = field(default_factory=dict)

    def _apply(self, balances: List[Balance]):
        for balance in balances:
            self.balances[balance.asset] = balance

    def _update_free(self, asset: str, amount: Decimal):
        if asset not in self.balances:
            raise ValueError(
                f"Asset {asset} not found in balances. Please add it in MockConnectorConfig"
            )
        self.balances[asset].free += amount

    def _update_locked(self, asset: str, amount: Decimal):
        if asset not in self.balances:
            raise ValueError(
                f"Asset {asset} not found in balances. Please add it in MockConnectorConfig"
            )
        self.balances[asset].locked += amount

    @property
    def balance_total(self) -> Dict[str, Decimal]:
        return {asset: balance.total for asset, balance in self.balances.items()}

    @property
    def balance_free(self) -> Dict[str, Decimal]:
        return {asset: balance.free for asset, balance in self.balances.items()}

    @property
    def balance_locked(self) -> Dict[str, Decimal]:
        return {asset: balance.locked for asset, balance in self.balances.items()}


class Precision(Struct):
    """
     "precision": {
      "amount": 0.0001,
      "price": 1e-05,
      "cost": null,
      "base": 1e-08,
      "quote": 1e-08
    },
    """

    amount: float
    price: float
    cost: float | None = None
    base: float | None = None
    quote: float | None = None


class LimitMinMax(Struct):
    """
    "limits": {
      "amount": {
        "min": 0.0001,
        "max": 1000.0
      },
      "price": {
        "min": 1e-05,
        "max": 1000000.0
      },
      "cost": {
        "min": 0.01,
        "max": 1000000.0
      }
    },
    """

    min: float | None
    max: float | None


class Limit(Struct):
    leverage: LimitMinMax = None
    amount: LimitMinMax = None
    price: LimitMinMax = None
    cost: LimitMinMax = None
    market: LimitMinMax = None


class MarginMode(Struct):
    isolated: bool | None
    cross: bool | None


class BaseMarket(Struct):
    """Base market structure for all exchanges."""

    id: str
    lowercaseId: str | None
    symbol: str
    base: str
    quote: str
    settle: str | None
    baseId: str
    quoteId: str
    settleId: str | None
    type: InstrumentType
    spot: bool
    margin: bool | None
    swap: bool
    future: bool
    option: bool
    index: bool | str | None
    active: bool
    contract: bool
    linear: bool | None
    inverse: bool | None
    subType: InstrumentType | None
    taker: float
    maker: float
    contractSize: float | None
    expiry: int | None
    expiryDatetime: str | None
    strike: float | str | None
    optionType: str | None
    precision: Precision
    limits: Limit
    marginModes: MarginMode
    created: int | None
    tierBased: bool | None
    percentage: bool | None
    # feeSide: str  # not supported by okx exchanges


"""
class Position(Struct):

    one-way mode:
    > order (side: buy) -> side: buy | pos_side: net/both | reduce_only: False [open long position]
    > order (side: sell) -> side: sell | pos_side: net/both | reduce_only: False [open short position]
    > order (side: buy, reduce_only=True) -> side: buy | pos_side: net/both | reduce_only: True [close short position]
    > order (side: sell, reduce_only=True) -> side: sell | pos_side: net/both | reduce_only: True [close long position]

    hedge mode:
    > order (side: buy, pos_side: long) -> side: buy | pos_side: long | reduce_only: False [open long position]
    > order (side: sell, pos_side: short) -> side: sell | pos_side: short | reduce_only: False [open short position]
    > order (side: sell, pos_side: long) -> side: sell | pos_side: long | reduce_only: True [close long position]
    > order (side: buy, pos_side: short) -> side: buy | pos_side: short | reduce_only: True [close short position]

    
"""


class Position(Struct):
    symbol: str
    exchange: ExchangeType
    signed_amount: Decimal = Decimal("0")
    entry_price: float = 0
    side: Optional[PositionSide] = None
    unrealized_pnl: float = 0
    realized_pnl: float = 0

    @property
    def amount(self, contract_size: Decimal = Decimal("1")) -> Decimal:
        return abs(self.signed_amount) * contract_size

    @property
    def is_opened(self) -> bool:
        return self.amount != Decimal("0")

    @property
    def is_closed(self) -> bool:
        return not self.is_opened

    @property
    def is_long(self) -> bool:
        return self.side == PositionSide.LONG

    @property
    def is_short(self) -> bool:
        return self.side == PositionSide.SHORT


class KlineList(list[Kline]):
    def __init__(self, klines: list[Kline], fields: list[str] | None = None):
        super().__init__(klines)
        self._fields = fields or [
            "timestamp",
            "symbol",
            "open",
            "high",
            "low",
            "close",
            "volume",
            "confirm",
        ]
        # Validate that all fields exist in Kline
        for item in self._fields:
            if not hasattr(Kline, item) and item != "timestamp":
                raise ValueError(f"Field {item} does not exist in Kline")

    @property
    def df(self):
        data = {}
        for item in self._fields:
            if item == "timestamp":
                data[item] = [kline.start for kline in self]
            else:
                data[item] = [getattr(kline, item) for kline in self]

        df = pd.DataFrame(data)
        df["date"] = pd.to_datetime(df["timestamp"], unit="ms", utc=True)
        df.set_index("date", inplace=True)
        return df

    @property
    def values(self):
        return sorted(self, key=lambda x: x.timestamp)



"""
Base factory interface for exchange components.

This module defines the abstract base class and context for creating exchange-specific
components in a standardized way.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Optional
from nexustrader.constants import ExchangeType, AccountType
from nexustrader.base import (
    ExchangeManager,
    PublicConnector,
    PrivateConnector,
    ExecutionManagementSystem,
)
from nexustrader.config import (
    PublicConnectorConfig,
    PrivateConnectorConfig,
    BasicConfig,
)


@dataclass
class BuildContext:
    """Context object containing shared components for factory methods."""

    msgbus: Any
    clock: Any
    task_manager: Any
    cache: Any
    registry: Any  # OrderRegistry
    is_mock: bool


class ExchangeFactory(ABC):
    """
    Abstract base class for creating exchange-specific components.

    Each exchange should implement this factory to provide standardized
    creation methods for all its components.
    """

    @property
    @abstractmethod
    def exchange_type(self) -> ExchangeType:
        """Return the exchange type this factory handles."""
        pass

    @abstractmethod
    def create_manager(self, basic_config: BasicConfig) -> ExchangeManager:
        """
        Create an ExchangeManager instance.

        Args:
            basic_config: Basic configuration containing API credentials

        Returns:
            Configured ExchangeManager instance
        """
        pass

    @abstractmethod
    def create_public_connector(
        self,
        config: PublicConnectorConfig,
        exchange: ExchangeManager,
        context: BuildContext,
    ) -> PublicConnector:
        """
        Create a PublicConnector instance.

        Args:
            config: Public connector configuration
            exchange: Exchange manager instance
            context: Build context with shared components

        Returns:
            Configured PublicConnector instance
        """
        pass

    @abstractmethod
    def create_private_connector(
        self,
        config: PrivateConnectorConfig,
        exchange: ExchangeManager,
        context: BuildContext,
        account_type: Optional[AccountType] = None,
    ) -> PrivateConnector:
        """
        Create a PrivateConnector instance.

        Args:
            config: Private connector configuration
            exchange: Exchange manager instance
            context: Build context with shared components
            account_type: Override account type (for testnet handling)

        Returns:
            Configured PrivateConnector instance
        """
        pass

    @abstractmethod
    def create_ems(
        self, exchange: ExchangeManager, context: BuildContext
    ) -> ExecutionManagementSystem:
        """
        Create an ExecutionManagementSystem instance.

        Args:
            exchange: Exchange manager instance
            context: Build context with shared components

        Returns:
            Configured ExecutionManagementSystem instance
        """
        pass

    def validate_config(self, basic_config: BasicConfig) -> None:
        """
        Validate exchange-specific configuration.

        Args:
            basic_config: Basic configuration to validate

        Raises:
            ValueError: If configuration is invalid
        """
        # Default implementation - exchanges can override
        pass

    def get_private_account_type(
        self, exchange: ExchangeManager, config: PrivateConnectorConfig
    ) -> AccountType:
        """
        Determine the account type for private connector.

        Some exchanges (like Bybit, OKX) map testnet to different account types.

        Args:
            exchange: Exchange manager instance
            config: Private connector configuration

        Returns:
            Account type to use
        """
        return config.account_type

    def setup_public_connector(
        self,
        exchange: ExchangeManager,
        connector: PublicConnector,
        account_type: AccountType,
    ) -> None:
        """
        Perform any post-creation setup for public connector.

        Some exchanges (like Bitget) need additional setup after creation.

        Args:
            exchange: Exchange manager instance
            connector: Created public connector
            account_type: Account type being used
        """
        # Default implementation - exchanges can override
        pass



import os
import time
import threading
from typing import Dict, Any
from decimal import Decimal
import redis
import msgspec
import nexuslog as logging

from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import LiveClock
from nexustrader.core.entity import get_redis_client_if_available


class StrategyStateExporter:
    """Exports strategy state to files for CLI monitoring"""

    def __init__(
        self, strategy_id: str, user_id: str, cache: AsyncCache, clock: LiveClock
    ):
        self.strategy_id = strategy_id
        self.user_id = user_id
        self.cache = cache
        self.log = logging.getLogger(name=type(self).__name__)
        self.clock = clock

        # Setup Redis connection and strategy tracking
        self.full_id = f"{strategy_id}_{user_id}"
        self.pid = os.getpid()
        self.strategy_key = f"{self.full_id}:{self.pid}"
        self._redis_client = None
        self._init_redis()

        # Export control
        self.running = False
        self.export_thread = None
        self.heartbeat_thread = None

    def start(self):
        """Start state export"""
        if self.running:
            return

        self.log.debug(f"Starting state export for {self.full_id}")

        # Register strategy in Redis running set
        self._register_strategy()

        self.running = True

        # Start export thread
        self.export_thread = threading.Thread(target=self._export_loop, daemon=True)
        self.export_thread.start()

        # Start heartbeat thread (now just updates Redis TTL)
        self.heartbeat_thread = threading.Thread(
            target=self._heartbeat_loop, daemon=True
        )
        self.heartbeat_thread.start()

    def stop(self):
        """Stop state export"""
        if not self.running:
            return

        self.log.debug(f"Stopping state export for {self.full_id}")
        self.running = False

        # Wait for threads to finish
        if self.export_thread and self.export_thread.is_alive():
            self.export_thread.join(timeout=0.1)

        if self.heartbeat_thread and self.heartbeat_thread.is_alive():
            self.heartbeat_thread.join(timeout=0.1)

        # Remove strategy from Redis running set
        self._unregister_strategy()

    def _export_loop(self):
        """Main export loop"""
        while self.running:
            try:
                self._export_state()
                time.sleep(1)  # Export every 1 second
            except Exception as e:
                self.log.error(f"Error in export loop: {e}")
                time.sleep(1)  # Back off on error

    def _register_strategy(self):
        """Register strategy in Redis running set"""
        if not self._redis_client:
            self.log.warning("Redis client not available, cannot register strategy")
            return

        try:
            strategy_info = {
                "strategy_id": self.strategy_id,
                "user_id": self.user_id,
                "pid": self.pid,
                "start_time": self.clock.timestamp(),
                "full_id": self.full_id,
            }

            # Add to running strategies hash
            self._redis_client.hset(
                "nexustrader:running_strategies",
                self.strategy_key,
                msgspec.json.encode(strategy_info),
            )

            self.log.debug(f"Registered strategy {self.strategy_key} in Redis")
        except Exception as e:
            self.log.error(f"Failed to register strategy in Redis: {e}")

    def _unregister_strategy(self):
        """Remove strategy from Redis running set"""
        if not self._redis_client:
            return

        try:
            self._redis_client.hdel("nexustrader:running_strategies", self.strategy_key)
            self.log.debug(f"Unregistered strategy {self.strategy_key} from Redis")
        except Exception as e:
            self.log.error(f"Failed to unregister strategy from Redis: {e}")

    def _heartbeat_loop(self):
        """Background heartbeat loop - maintains Redis strategy info"""
        while self.running:
            try:
                if self._redis_client:
                    # Update strategy info with current timestamp
                    strategy_info = {
                        "strategy_id": self.strategy_id,
                        "user_id": self.user_id,
                        "pid": self.pid,
                        "start_time": self.clock.timestamp(),
                        "full_id": self.full_id,
                        "last_heartbeat": self.clock.timestamp(),
                    }

                    self._redis_client.hset(
                        "nexustrader:running_strategies",
                        self.strategy_key,
                        msgspec.json.encode(strategy_info),
                    )

                time.sleep(30)  # Update every 30 seconds
            except Exception as e:
                self.log.debug(f"Heartbeat failed: {e}")
                time.sleep(5)  # Retry more frequently if failed

    def _export_state(self):
        """Export current strategy state"""
        timestamp = self.clock.timestamp()

        # Export positions
        self._export_positions(timestamp)

        # Export balances
        self._export_balances(timestamp)

        # Export metrics
        self._export_metrics(timestamp)

    def _export_positions(self, timestamp: float):
        """Export positions state"""
        positions_data = {"timestamp": timestamp, "positions": []}

        for symbol, position in self.cache._mem_positions.items():
            positions_data["positions"].append(
                {
                    "symbol": position.symbol,
                    "side": position.side.value if position.side else None,
                    "amount": float(position.amount),
                    "entry_price": position.entry_price,
                }
            )

        self._save_to_redis("positions", positions_data)

    def _export_balances(self, timestamp: float):
        """Export balances state"""
        balances_data = {"timestamp": timestamp, "balances": []}

        for account_type, account_balance in self.cache._mem_account_balance.items():
            for asset, balance in account_balance.balances.items():
                balances_data["balances"].append(
                    {
                        "account_type": account_type.value if account_type else None,
                        "asset": asset,
                        "free": float(balance.free) if balance.free else None,
                        "locked": float(balance.locked) if balance.locked else None,
                        "total": float(balance.total) if balance.total else None,
                    }
                )

        self._save_to_redis("balances", balances_data)

    def _export_metrics(self, timestamp: float):
        """Export performance metrics"""
        metrics_data = {
            "timestamp": timestamp,
            "metrics": {
                "total_orders": len(self.cache._mem_orders),
                "open_orders": len(
                    [
                        o
                        for o in self.cache._mem_orders.values()
                        if o.status and o.status.value in ["NEW", "PARTIALLY_FILLED"]
                    ]
                ),
                "total_positions": len(self.cache._mem_positions),
                "active_positions": len(
                    [
                        p
                        for p in self.cache._mem_positions.values()
                        if p.amount and p.amount > Decimal("0")
                    ]
                ),
            },
        }

        # Calculate total unrealized PnL
        total_unrealized_pnl = sum(
            float(pos.unrealized_pnl) if pos.unrealized_pnl else 0.0
            for pos in self.cache._mem_positions.values()
        )
        metrics_data["metrics"]["total_unrealized_pnl"] = total_unrealized_pnl

        # Calculate total realized PnL
        total_realized_pnl = sum(
            float(pos.realized_pnl) if pos.realized_pnl else 0.0
            for pos in self.cache._mem_positions.values()
        )
        metrics_data["metrics"]["total_realized_pnl"] = total_realized_pnl

        self._save_to_redis("metrics", metrics_data)

    def _init_redis(self):
        """Initialize Redis connection"""
        self._redis_client = get_redis_client_if_available()
        if self._redis_client:
            self.log.debug("Redis connection established")
        else:
            self.log.debug("Redis not available")
            self._redis_client = None

    def _save_to_redis(self, data_type: str, data: Dict[str, Any]):
        """Save data to Redis using msgspec"""
        if not self._redis_client:
            self.log.warning("Redis client not available, skipping save")
            return

        try:
            # Serialize with msgspec
            json_data = msgspec.json.encode(data)
            key = f"strategy:{self.strategy_key}:{data_type}"
            self._redis_client.set(key, json_data, ex=300)  # 5 minute expiry
            self.log.debug(f"Saved {data_type} to Redis key: {key}")
        except (redis.RedisError, redis.ConnectionError) as e:
            self.log.error(f"Redis error saving {data_type}: {e}")
        except msgspec.EncodeError as e:
            self.log.error(f"Serialization error for {data_type}: {e}")



from typing import Callable, List, Dict
from typing import Any
from urllib.parse import urlencode
from nexustrader.base import WSClient
from nexustrader.exchange.binance.constants import (
    BinanceAccountType,
    BinanceKlineInterval,
    BinanceRateLimiter,
)
from nexustrader.core.entity import TaskManager
from nexustrader.core.nautilius_core import hmac_signature, LiveClock


class BinanceWSClient(WSClient):
    def __init__(
        self,
        account_type: BinanceAccountType,
        handler: Callable[..., Any],
        task_manager: TaskManager,
        clock: LiveClock,
        ws_suffix: str = "/ws",
        custom_url: str | None = None,
        max_subscriptions_per_client: int | None = None,
        max_clients: int | None = None,
    ):
        self._account_type = account_type
        url = account_type.ws_url

        if ws_suffix not in ["/ws", "/stream"]:
            raise ValueError(f"Invalid ws_suffix: {ws_suffix}")

        url += ws_suffix

        if custom_url is not None:
            url = custom_url

        super().__init__(
            url,
            handler=handler,
            task_manager=task_manager,
            clock=clock,
            enable_auto_ping=False,
            max_subscriptions_per_client=max_subscriptions_per_client,
            max_clients=max_clients,
        )

    def _send_payload(
        self,
        params: List[str],
        method: str = "SUBSCRIBE",
        chunk_size: int = 50,
        client_id: int | None = None,
    ):
        # Split params into chunks of 100 if length exceeds 100
        params_chunks = [
            params[i : i + chunk_size] for i in range(0, len(params), chunk_size)
        ]

        for chunk in params_chunks:
            payload = {
                "method": method,
                "params": chunk,
                "id": self._clock.timestamp_ms(),
            }
            self.send(payload, client_id=client_id)

    def _subscribe(self, params: List[str]):
        assigned = self._register_subscriptions(params)
        if not assigned:
            return
        for client_id, client_params in assigned.items():
            for param in client_params:
                self._log.debug(f"Subscribing to {param}...")
            if self._is_client_connected(client_id):
                self._send_payload(
                    client_params, method="SUBSCRIBE", client_id=client_id
                )

    def _unsubscribe(self, params: List[str]):
        removed = self._unregister_subscriptions(params)
        if not removed:
            return
        for client_id, client_params in removed.items():
            for param in client_params:
                self._log.debug(f"Unsubscribing from {param}...")
            self._send_payload(client_params, method="UNSUBSCRIBE", client_id=client_id)

    def subscribe_agg_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@aggTrade" for symbol in symbols]
        self._subscribe(params)

    def subscribe_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@trade" for symbol in symbols]
        self._subscribe(params)

    def subscribe_book_ticker(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@bookTicker" for symbol in symbols]
        self._subscribe(params)

    def subscribe_partial_book_depth(self, symbols: List[str], level: int):
        if level not in (5, 10, 20):
            raise ValueError("Level must be 5, 10, or 20")
        params = [f"{symbol.lower()}@depth{level}@100ms" for symbol in symbols]
        self._subscribe(params)

    def subscribe_mark_price(self, symbols: List[str]):
        if not self._account_type.is_future:
            raise ValueError("Only Supported for `Future Account`")
        params = [f"{symbol.lower()}@markPrice@1s" for symbol in symbols]
        self._subscribe(params)

    def subscribe_user_data_stream(self, listen_key: str):
        self._subscribe([listen_key])

    def subscribe_kline(
        self,
        symbols: List[str],
        interval: BinanceKlineInterval,
    ):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@kline_{interval.value}" for symbol in symbols]
        self._subscribe(params)

    def unsubscribe_agg_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@aggTrade" for symbol in symbols]
        self._unsubscribe(params)

    def unsubscribe_trade(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@trade" for symbol in symbols]
        self._unsubscribe(params)

    def unsubscribe_book_ticker(self, symbols: List[str]):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@bookTicker" for symbol in symbols]
        self._unsubscribe(params)

    def unsubscribe_partial_book_depth(self, symbols: List[str], level: int):
        if level not in (5, 10, 20):
            raise ValueError("Level must be 5, 10, or 20")
        params = [f"{symbol.lower()}@depth{level}@100ms" for symbol in symbols]
        self._unsubscribe(params)

    def unsubscribe_mark_price(self, symbols: List[str]):
        if not self._account_type.is_future:
            raise ValueError("Only Supported for `Future Account`")
        params = [f"{symbol.lower()}@markPrice@1s" for symbol in symbols]
        self._unsubscribe(params)

    def unsubscribe_user_data_stream(self, listen_key: str):
        self._unsubscribe([listen_key])

    def unsubscribe_kline(
        self,
        symbols: List[str],
        interval: BinanceKlineInterval,
    ):
        if (
            self._account_type.is_isolated_margin_or_margin
            or self._account_type.is_portfolio_margin
        ):
            raise ValueError(
                "Not Supported for `Margin Account` or `Portfolio Margin Account`"
            )
        params = [f"{symbol.lower()}@kline_{interval.value}" for symbol in symbols]
        self._unsubscribe(params)

    async def _resubscribe_for_client(self, client_id: int, subscriptions: List[str]):
        if not subscriptions:
            return
        self._send_payload(subscriptions, client_id=client_id)


class BinanceWSApiClient(WSClient):
    def __init__(
        self,
        account_type: BinanceAccountType,
        api_key: str,
        secret: str,
        handler: Callable[..., Any],
        task_manager: TaskManager,
        clock: LiveClock,
        enable_rate_limit: bool,
    ):
        self._account_type = account_type
        self._api_key = api_key
        self._secret = secret
        self._limiter = BinanceRateLimiter(enable_rate_limit)

        url = account_type.ws_order_url

        if not url:
            raise ValueError(f"WebSocket URL not supported for {account_type}")

        super().__init__(
            url=url,
            handler=handler,
            task_manager=task_manager,
            clock=clock,
            enable_auto_ping=False,
        )

    def _generate_signature_v2(self, query: str) -> str:
        signature = hmac_signature(self._secret, query)
        return signature

    def _send_payload(
        self,
        id: str,
        method: str,
        params: Dict[str, Any],
        required_ts: bool = True,
        auth: bool = True,
    ):
        if required_ts:
            params["timestamp"] = self._clock.timestamp_ms()

        if auth:
            params["apiKey"] = self._api_key
            query = urlencode(sorted(params.items()))
            signature = self._generate_signature_v2(query)
            params["signature"] = signature

        payload = {
            "method": method,
            "id": id,
            "params": params,
        }
        self.send(payload)

    async def spot_new_order(
        self, oid: str, symbol: str, side: str, type: str, quantity: str, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            **kwargs,
        }
        await self._limiter.api_order_limit(cost=1)
        self._send_payload(id=f"n{oid}", method="order.place", params=params)

    async def spot_cancel_order(
        self, oid: str, symbol: str, origClientOrderId: int, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "origClientOrderId": origClientOrderId,
            **kwargs,
        }
        await self._limiter.api_order_limit(cost=1)
        self._send_payload(id=f"c{oid}", method="order.cancel", params=params)

    async def usdm_new_order(
        self, oid: str, symbol: str, side: str, type: str, quantity: str, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            **kwargs,
        }
        await self._limiter.fapi_order_limit(cost=0)
        self._send_payload(id=f"n{oid}", method="order.place", params=params)

    async def usdm_cancel_order(
        self, oid: str, symbol: str, origClientOrderId: int, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "origClientOrderId": origClientOrderId,
            **kwargs,
        }
        await self._limiter.fapi_order_limit(cost=1)
        self._send_payload(id=f"c{oid}", method="order.cancel", params=params)

    async def coinm_new_order(
        self, oid: str, symbol: str, side: str, type: str, quantity: str, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            **kwargs,
        }
        await self._limiter.dapi_order_limit(cost=0)
        self._send_payload(id=f"n{oid}", method="order.place", params=params)

    async def coinm_cancel_order(
        self, oid: str, symbol: str, origClientOrderId: int, **kwargs: Any
    ):
        params = {
            "symbol": symbol,
            "origClientOrderId": origClientOrderId,
            **kwargs,
        }
        await self._limiter.dapi_order_limit(cost=1)
        self._send_payload(id=f"c{oid}", method="order.cancel", params=params)

    def _subscribe(self, subscriptions: List[tuple[Dict[str, Any], str]]):
        assigned = self._register_subscriptions(subscriptions)
        if not assigned:
            return
        for client_id, client_subs in assigned.items():
            for params, method in client_subs:
                self._log.debug(f"Subscribing to {method} with {params}...")
            if self._is_client_connected(client_id):
                for params, method in client_subs:
                    request_id = f"{self._clock.timestamp_ms()}"
                    self._send_payload(
                        id=request_id,
                        method=method,
                        params=dict(params),
                        auth=True,
                    )

    async def _resubscribe_for_client(
        self, client_id: int, subscriptions: List[tuple[Dict[str, Any], str]]
    ):
        if not subscriptions:
            return
        for params, method in subscriptions:
            request_id = f"s{self._clock.timestamp_ms()}"
            self._send_payload(
                id=request_id,
                method=method,
                params=dict(params),
                auth=True,
            )

    def subscribe_spot_user_data_stream(self):
        if not self._account_type.is_spot:
            raise ValueError("Only Supported for `Spot Account`")
        params: Dict[str, Any] = {}
        self._subscribe([(params, "userDataStream.subscribe.signature")])



from typing import Optional, Dict
import nexuslog as logging

from nexustrader.schema import Order


class OrderRegistry:
    def __init__(self):
        self._log = logging.getLogger(name=type(self).__name__)
        self._tmp_order: Dict[str, Order] = {}
        self._oids: set[str] = set()

    def register_order(self, oid: str) -> None:
        """Register an order to track its status"""
        self._log.debug(f"[ORDER REGISTER]: {oid}")
        self._oids.add(oid)

    def is_registered(self, oid: str) -> bool:
        """Check if an order is registered"""
        return oid in self._oids

    def unregister_order(self, oid: str) -> None:
        """Remove order mapping when no longer needed"""
        self._log.debug(f"[ORDER UNREGISTER]: {oid}")
        self._oids.discard(oid)

    def register_tmp_order(self, order: Order) -> None:
        """Register a temporary order"""
        self._tmp_order[order.oid] = order
        self._log.debug(f"[TMP ORDER REGISTER]: {order.oid}")

    def unregister_tmp_order(self, oid: str) -> None:
        """Unregister a temporary order"""
        self._log.debug(f"[TMP ORDER UNREGISTER]: {oid}")
        self._tmp_order.pop(oid, None)

    def get_tmp_order(self, oid: str) -> Optional[Order]:
        self._log.debug(f"[TMP ORDER GET]: {oid}")
        return self._tmp_order.get(oid, None)



from dataclasses import dataclass, field
from typing import Dict, List, Literal
from nexustrader.constants import AccountType, ExchangeType, StorageType
from nexustrader.strategy import Strategy
from zmq.asyncio import Socket

LOG_LEVELS = Literal["TRACE", "DEBUG", "INFO", "WARNING", "ERROR"]


@dataclass
class LogConfig:
    """LogConfig for NexusTrader logging system using nexuslog.

    Maps directly to nexuslog.basicConfig() parameters.

    Attributes:
        filename: Optional file path for log output. If None, logs to stdout.
        level: Minimum log level to record. Options: TRACE, DEBUG, INFO, WARNING, ERROR
        unix_ts: If True, emit unix timestamps instead of formatted local time.
    """

    filename: str | None = None
    level: LOG_LEVELS = "INFO"
    name_levels: dict[str | None, LOG_LEVELS] | None = None
    unix_ts: bool = False
    batch_size: int = 1

    def __post_init__(self):
        if self.level not in ["TRACE", "DEBUG", "INFO", "WARNING", "ERROR"]:
            raise ValueError(
                f"Invalid level: {self.level}. Must be one of TRACE, DEBUG, INFO, WARNING, ERROR."
            )

        if self.name_levels:
            for name, level in self.name_levels.items():
                if level not in ["TRACE", "DEBUG", "INFO", "WARNING", "ERROR"]:
                    raise ValueError(
                        f"Invalid level: {level} for logger name: {name}. Must be one of TRACE, DEBUG, INFO, WARNING, ERROR."
                    )


@dataclass
class BasicConfig:
    """
    Basic configuration for the trading system.

    Attributes:
        api_key (str): API key for the exchange. For HyperLiquid, this is the real wallet address, not the address generated by the Agent.
        secret (str): Secret key for the exchange. For HyperLiquid, this is the Agent private key, you can get it at https://app.hyperliquid-testnet.xyz/API
        testnet (bool): Whether to use the testnet environment. Defaults to False.
        passphrase (str | None): Passphrase for the exchange, if required. Defaults to None.
    """

    api_key: str | None = None
    secret: str | None = None
    testnet: bool = False
    passphrase: str = None


@dataclass
class PublicConnectorConfig:
    account_type: AccountType
    enable_rate_limit: bool = True
    custom_url: str | None = None
    max_subscriptions_per_client: int | None = None
    max_clients: int | None = None


@dataclass
class PrivateConnectorConfig:
    """
    Configuration for private connector settings.

    Attributes
    ----------
    account_type : AccountType
        The type of account for the private connector.
    enable_rate_limit : bool, default=True
        Whether to enable rate limiting for API requests.
    max_retries for failed order request : int, default=0
    delay_initial_ms : int, default=100
    delay_max_ms : int, default=800
    backoff_factor : int, default=2
    max_slippage : float, default=0.02
        Maximum slippage allowed for market orders, expressed as a percentage (e.g., 0.02 for 2%). Only applicable for Bitget and HyperLiquid exchanges.
    """

    account_type: AccountType
    enable_rate_limit: bool = True
    max_retries: int = 0
    delay_initial_ms: int = 100
    delay_max_ms: int = 800
    backoff_factor: int = 2
    max_slippage: float = 0.02  # 2% slippage
    max_subscriptions_per_client: int | None = None
    max_clients: int | None = None


@dataclass
class ZeroMQSignalConfig:
    """ZeroMQ Signal Configuration Class.

    Used to configure the ZeroMQ subscriber socket to receive custom trade signals.

    Attributes:
        socket (`zmq.asyncio.Socket`): ZeroMQ asynchronous socket object

    Example:
        >>> from zmq.asyncio import Context
        >>> context = Context()
        >>> socket = context.socket(zmq.SUB)
        >>> socket.connect("ipc:///tmp/zmq_custom_signal")
        >>> socket.setsockopt(zmq.SUBSCRIBE, b"")
        >>> config = ZeroMQSignalConfig(socket=socket)
    """

    socket: Socket


@dataclass
class WebConfig:
    """Configuration for the optional FastAPI web server."""

    enabled: bool = False
    host: str = "0.0.0.0"
    port: int = 8000
    log_level: Literal["critical", "error", "warning", "info", "debug"] = "info"

    def __post_init__(self):
        if not (0 < self.port <= 65535):
            raise ValueError("port must be between 1 and 65535")


@dataclass
class MockConnectorConfig:
    initial_balance: Dict[str, float | int]
    account_type: AccountType
    fee_rate: float = 0.0005
    quote_currency: str = "USDT"
    overwrite_balance: bool = False
    overwrite_position: bool = False
    update_interval: int = 60
    leverage: float = 1.0

    def __post_init__(self):
        if not self.account_type.is_mock:
            raise ValueError(
                f"Invalid account type: {self.account_type} for mock connector. Must be `LINEAR_MOCK`, `INVERSE_MOCK`, or `SPOT_MOCK`."
            )


@dataclass
class Config:
    """
    Represents the configuration for a trading system.

    This class holds all necessary configuration information including strategy details,
    exchange configurations, connector setups, database settings, and logging preferences.

    Attributes:
        strategy_id (str): Unique identifier for the strategy.
        user_id (str): Identifier for the user running the strategy.
        strategy (Strategy): The trading strategy to be executed.
        basic_config (Dict[ExchangeType, BasicConfig]): Basic configuration for each exchange.
        public_conn_config (Dict[ExchangeType, List[PublicConnectorConfig]]): Public connector configurations by exchange.
        private_conn_config (Dict[ExchangeType, List[PrivateConnectorConfig | MockConnectorConfig]]):
            Private connector configurations by exchange, can include mock connectors.
        zero_mq_signal_config (ZeroMQSignalConfig | None): Configuration for ZeroMQ signal, if used.
        db_path (str): Path to the database file. Defaults to ".keys/cache.db".
        storage_backend (StorageType): Type of storage backend to use. Defaults to SQLITE.
        cache_sync_interval (int): Interval in seconds for cache synchronization. Defaults to 60.
        cache_expired_time (int): Time in seconds after which cache entries expire. Defaults to 3600.
        is_mock (bool): Flag indicating if the system is running in mock mode. Defaults to False.
        log_config (LogConfig): Configuration for logging. Defaults to a new LogConfig instance.
        enable_cli (bool): Flag to enable command-line interface. Defaults to False.
        web_config (WebConfig): Settings for the optional FastAPI web interface.
        flashduty_integration_key (str | None): FlashDuty integration key for push alerts. Defaults to None.
        exit_after_cancel (bool): Whether to cancel all open orders when the engine is disposed (e.g., on Ctrl+C). Defaults to True.

    Notes:
        The __post_init__ method enforces that you cannot mix mock and real private connectors.
        Either all private connectors must be mock or all must be real.
    """

    strategy_id: str
    user_id: str
    strategy: Strategy
    basic_config: Dict[ExchangeType, BasicConfig]
    public_conn_config: Dict[ExchangeType, List[PublicConnectorConfig]]
    private_conn_config: Dict[
        ExchangeType, List[PrivateConnectorConfig | MockConnectorConfig]
    ] = field(default_factory=dict)
    zero_mq_signal_config: ZeroMQSignalConfig | None = None
    db_path: str = ".keys/cache.db"
    storage_backend: StorageType = StorageType.SQLITE
    cache_sync_interval: int = 60
    cache_expired_time: int = 3600
    is_mock: bool = False
    log_config: LogConfig = field(default_factory=LogConfig)
    enable_cli: bool = False
    web_config: WebConfig = field(default_factory=WebConfig)
    exit_after_cancel: bool = True
    flashduty_integration_key: str | None = None

    def __post_init__(self):
        # Check if any connector is mock, then all must be mock
        has_mock = False
        has_private = False

        for connectors in self.private_conn_config.values():
            for connector in connectors:
                if isinstance(connector, MockConnectorConfig):
                    has_mock = True
                elif isinstance(connector, PrivateConnectorConfig):
                    has_private = True

                if has_mock and has_private:
                    raise ValueError(
                        "Cannot mix mock and real private connectors. Use either all mock or all private connectors."
                    )

        self.is_mock = has_mock



import os
import sys
from typing import Literal, Dict, List, TypedDict, NotRequired
from enum import Enum
from dynaconf import Dynaconf

BACKEND_LITERAL = Literal["memory", "redis"]


def is_sphinx_build():
    return "sphinx" in sys.modules


if not os.path.exists(".keys/"):
    os.makedirs(".keys/")
if not os.path.exists(".keys/.secrets.toml") and not is_sphinx_build():
    raise FileNotFoundError(
        "Config file not found, please create a config file at .keys/.secrets.toml"
    )


settings = Dynaconf(
    envvar_prefix="NEXUS",
    settings_files=[".keys/settings.toml", ".keys/.secrets.toml"],
    warn_dynaconf_global_settings=True,
    environments=False,
    load_dotenv=True,
)


def get_postgresql_config():
    """Get PostgreSQL configuration from environment variables or settings files."""
    return {
        "host": settings.get("PG_HOST", "localhost"),
        "port": settings.get("PG_PORT", 5432),
        "user": settings.get("PG_USER", "postgres"),
        "password": settings.get("PG_PASSWORD", ""),
        "database": settings.get("PG_DATABASE", "postgres"),
    }


def get_redis_config():
    try:
        return {
            "host": settings.REDIS_HOST,
            "port": settings.REDIS_PORT,
            "db": settings.REDIS_DB,
            "password": settings.REDIS_PASSWORD,
        }
    except Exception as e:
        raise ValueError(f"Failed to get Redis password: {e}")


IntervalType = Literal[
    "1s",
    "1m",
    "3m",
    "5m",
    "15m",
    "30m",
    "1h",
    "2h",
    "4h",
    "6h",
    "8h",
    "12h",
    "1d",
    "3d",
    "1w",
    "1M",
]


class BookLevel(Enum):
    L5 = "5"
    L10 = "10"
    L20 = "20"
    L50 = "50"


class KlineInterval(Enum):
    SECOND_1 = "1s"
    MINUTE_1 = "1m"
    MINUTE_3 = "3m"
    MINUTE_5 = "5m"
    MINUTE_15 = "15m"
    MINUTE_30 = "30m"
    HOUR_1 = "1h"
    HOUR_2 = "2h"
    HOUR_4 = "4h"
    HOUR_6 = "6h"
    HOUR_8 = "8h"
    HOUR_12 = "12h"
    DAY_1 = "1d"
    DAY_3 = "3d"
    WEEK_1 = "1w"
    MONTH_1 = "1M"
    VOLUME = "volume"

    @property
    def seconds(self) -> int:
        return {
            KlineInterval.SECOND_1: 1,
            KlineInterval.MINUTE_1: 60,
            KlineInterval.MINUTE_3: 180,
            KlineInterval.MINUTE_5: 300,
            KlineInterval.MINUTE_15: 900,
            KlineInterval.MINUTE_30: 1800,
            KlineInterval.HOUR_1: 3600,
            KlineInterval.HOUR_2: 7200,
            KlineInterval.HOUR_4: 14400,
            KlineInterval.HOUR_6: 21600,
            KlineInterval.HOUR_8: 28800,
            KlineInterval.HOUR_12: 43200,
            KlineInterval.DAY_1: 86400,
            KlineInterval.DAY_3: 259200,
            KlineInterval.WEEK_1: 604800,
            KlineInterval.MONTH_1: 2592000,
            KlineInterval.VOLUME: 0,
        }[self]

    @property
    def nanoseconds(self) -> int:
        return self.seconds * 1_000_000_000

    @property
    def microseconds(self) -> int:
        return self.seconds * 1_000_000

    @property
    def milliseconds(self) -> int:
        return self.seconds * 1_000


class SubmitType(Enum):
    CREATE = 0
    CANCEL = 1
    TWAP = 2
    CANCEL_TWAP = 3
    VWAP = 4
    CANCEL_VWAP = 5
    TAKE_PROFIT_AND_STOP_LOSS = 6
    MODIFY = 7
    CANCEL_ALL = 8
    BATCH = 9
    CREATE_WS = 10
    CANCEL_WS = 11


class EventType(Enum):
    BOOKL1 = 0
    TRADE = 1
    KLINE = 2
    MARK_PRICE = 3
    FUNDING_RATE = 4
    INDEX_PRICE = 5


class AlgoOrderStatus(Enum):
    RUNNING = "RUNNING"
    CANCELING = "CANCELING"
    FINISHED = "FINISHED"
    CANCELED = "CANCELED"
    FAILED = "FAILED"


class OrderStatus(Enum):
    # LOCAL
    INITIALIZED = "INITIALIZED"
    FAILED = "FAILED"
    CANCEL_FAILED = "CANCEL_FAILED"

    # IN-FLOW
    PENDING = "PENDING"
    CANCELING = "CANCELING"

    # OPEN
    ACCEPTED = "ACCEPTED"
    PARTIALLY_FILLED = "PARTIALLY_FILLED"

    # CLOSED
    FILLED = "FILLED"
    CANCELED = "CANCELED"
    EXPIRED = "EXPIRED"


class ExchangeType(Enum):
    BINANCE = "binance"
    OKX = "okx"
    BYBIT = "bybit"
    HYPERLIQUID = "hyperliquid"
    BITGET = "bitget"


class AccountType(Enum):
    pass


class OrderType(Enum):
    MARKET = "MARKET"
    LIMIT = "LIMIT"
    TAKE_PROFIT_MARKET = "TAKE_PROFIT_MARKET"
    TAKE_PROFIT_LIMIT = "TAKE_PROFIT_LIMIT"
    STOP_LOSS_MARKET = "STOP_LOSS_MARKET"
    STOP_LOSS_LIMIT = "STOP_LOSS_LIMIT"
    POST_ONLY = "POST_ONLY"

    @property
    def is_take_profit(self) -> bool:
        return self in (OrderType.TAKE_PROFIT_MARKET, OrderType.TAKE_PROFIT_LIMIT)

    @property
    def is_stop_loss(self) -> bool:
        return self in (OrderType.STOP_LOSS_MARKET, OrderType.STOP_LOSS_LIMIT)

    @property
    def is_market(self) -> bool:
        return self in (
            OrderType.MARKET,
            OrderType.TAKE_PROFIT_MARKET,
            OrderType.STOP_LOSS_MARKET,
        )

    @property
    def is_limit(self) -> bool:
        return self in (
            OrderType.LIMIT,
            OrderType.TAKE_PROFIT_LIMIT,
            OrderType.STOP_LOSS_LIMIT,
        )

    @property
    def is_post_only(self) -> bool:
        return self == OrderType.POST_ONLY


class TriggerType(Enum):
    LAST_PRICE = "LAST_PRICE"
    MARK_PRICE = "MARK_PRICE"
    INDEX_PRICE = "INDEX_PRICE"


class OrderSide(Enum):
    BUY = "BUY"
    SELL = "SELL"

    @property
    def is_buy(self) -> bool:
        return self == OrderSide.BUY

    @property
    def is_sell(self) -> bool:
        return self == OrderSide.SELL


class TimeInForce(Enum):
    GTC = "GTC"
    IOC = "IOC"
    FOK = "FOK"


class PositionSide(Enum):
    LONG = "LONG"
    SHORT = "SHORT"
    FLAT = "FLAT"

    @property
    def is_long(self) -> bool:
        return self == PositionSide.LONG

    @property
    def is_short(self) -> bool:
        return self == PositionSide.SHORT

    @property
    def is_flat(self) -> bool:
        return self == PositionSide.FLAT


class InstrumentType(Enum):
    SPOT = "spot"
    MARGIN = "margin"
    FUTURE = "future"
    OPTION = "option"
    SWAP = "swap"
    LINEAR = "linear"
    INVERSE = "inverse"


class OptionType(Enum):
    CALL = "call"
    PUT = "put"


STATUS_TRANSITIONS: Dict[OrderStatus, List[OrderStatus]] = {
    OrderStatus.PENDING: [
        OrderStatus.PENDING,
        OrderStatus.CANCELED,
        OrderStatus.CANCELING,
        OrderStatus.ACCEPTED,
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCEL_FAILED,
        OrderStatus.FAILED,  # e.g., rejected by exchange we mark as failed
    ],
    OrderStatus.CANCELING: [
        OrderStatus.CANCELED,
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
    ],
    OrderStatus.ACCEPTED: [
        OrderStatus.ACCEPTED,  # for modify order, pending -> accepted -> accepted -> (not allowed) -> pending
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCELING,
        OrderStatus.CANCELED,
        OrderStatus.EXPIRED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.PARTIALLY_FILLED: [
        OrderStatus.ACCEPTED,  # for modify order, accepted -> partially filled -> accepted -> (not allowed) -> pending
        OrderStatus.PARTIALLY_FILLED,
        OrderStatus.FILLED,
        OrderStatus.CANCELING,
        OrderStatus.CANCELED,
        OrderStatus.EXPIRED,
        OrderStatus.CANCEL_FAILED,
    ],
    OrderStatus.CANCEL_FAILED: [OrderStatus.FILLED],
    OrderStatus.FILLED: [],
    OrderStatus.CANCELED: [],
    OrderStatus.EXPIRED: [],
    OrderStatus.FAILED: [],
}


class DataType(Enum):
    BOOKL1 = "bookl1"
    BOOKL2 = "bookl2"
    TRADE = "trade"
    KLINE = "kline"
    VOLUME_KLINE = "volume_kline"
    MARK_PRICE = "mark_price"
    FUNDING_RATE = "funding_rate"
    INDEX_PRICE = "index_price"


class StorageType(Enum):
    SQLITE = "sqlite"
    POSTGRESQL = "postgresql"


class ParamBackend(Enum):
    MEMORY = "memory"
    REDIS = "redis"


class RateLimiter:
    pass


class RateLimiterSync:
    pass


class ConfigType(TypedDict):
    apiKey: str
    secret: str
    exchange_id: str
    sandbox: bool
    password: NotRequired[str]



import signal
import asyncio
import uuid
import warnings
import nexuslog as logging

from typing import Callable, Coroutine, Any, TypeVar, Union, cast
from typing import Dict, List
from dataclasses import dataclass
from nexustrader.core.nautilius_core import LiveClock
from nexustrader.schema import (
    Kline,
    BookL1,
    Trade,
    FundingRate,
    BookL2,
    IndexPrice,
    MarkPrice,
)

T = TypeVar("T")
InputDataType = Union[
    Kline, BookL1, Trade, FundingRate, BookL2, IndexPrice, IndexPrice, MarkPrice
]


class OidGen:
    __slots__ = ("_shard", "_last_ms", "_seq", "_clock")

    def __init__(self, clock: LiveClock):
        self._shard = self._generate_shard()
        self._last_ms = 0
        self._seq = 0
        self._clock = clock

    def _generate_shard(self) -> int:
        uuid_int = cast(int, uuid.uuid4().int)
        shard = (uuid_int % 9999) + 1
        return shard

    @property
    def oid(self) -> str:
        ms = self._clock.timestamp_ms()
        if ms == self._last_ms:
            self._seq = (self._seq + 1) % 1000
            if self._seq == 0:
                while True:
                    ms2 = self._clock.timestamp_ms()
                    if ms2 > ms:
                        ms = ms2
                        break
        else:
            self._seq = 0
        self._last_ms = ms
        return f"{ms:013d}{self._seq:03d}{self._shard:04d}"

    def get_shard(self) -> int:
        return self._shard


def is_redis_available() -> bool:
    """Check if Redis dependencies and server are available"""
    try:
        # Check if Redis Python packages are installed
        import redis
        import socket
        from nexustrader.constants import get_redis_config

        # Check if Redis server is accessible
        try:
            # Get Redis config and test connection
            redis_config = get_redis_config()
            client = redis.Redis(**redis_config)
            client.ping()  # This will raise an exception if Redis is not accessible
            client.close()
            return True
        except (redis.ConnectionError, redis.TimeoutError, socket.error, Exception):
            return False

    except ImportError:
        return False


def get_redis_client_if_available():
    import redis

    """Get Redis client if available, otherwise return None"""
    if not is_redis_available():
        return None

    try:
        from nexustrader.constants import get_redis_config

        redis_config = get_redis_config()
        return redis.Redis(**redis_config)
    except Exception:
        return None


@dataclass
class RateLimit:
    """
    max_rate: Allow up to max_rate / time_period acquisitions before blocking.
    time_period: Time period in seconds.
    """

    max_rate: float
    time_period: float = 60


class TaskManager:
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        enable_signal_handlers: bool = True,
        cancel_timeout: float | None = 5.0,
    ):
        self._log = logging.getLogger(name=type(self).__name__)
        self._tasks: Dict[str, asyncio.Task] = {}
        self._shutdown_event = asyncio.Event()
        self._loop = loop
        self._cancel_timeout = cancel_timeout
        if enable_signal_handlers:
            self._setup_signal_handlers()

    def _setup_signal_handlers(self):
        try:
            for sig in (signal.SIGINT, signal.SIGTERM):
                self._loop.add_signal_handler(
                    sig, lambda: self.create_task(self._shutdown())
                )
        except NotImplementedError:
            warnings.warn("Signal handlers not supported on this platform")

    async def _shutdown(self):
        self._shutdown_event.set()
        self._log.debug("Shutdown signal received, cleaning up...")

    def create_task(
        self, coro: Coroutine[Any, Any, Any], name: str | None = None
    ) -> asyncio.Task:
        task = asyncio.create_task(coro, name=name)
        self._tasks[task.get_name()] = task
        task.add_done_callback(self._handle_task_done)
        return task

    def run_sync(self, coro: Coroutine[Any, Any, T]) -> T:
        """
        Run an async coroutine in a synchronous context.

        Args:
            coro: The coroutine to run

        Returns:
            The result of the coroutine

        Raises:
            RuntimeError: If the event loop is not running and cannot be started
            Exception: Any exception raised by the coroutine
        """
        try:
            if self._loop.is_running():
                future = asyncio.run_coroutine_threadsafe(coro, self._loop)
                return future.result()
            else:
                if self._loop.is_closed():
                    raise RuntimeError("Event loop is closed")
                return self._loop.run_until_complete(coro)
        except asyncio.CancelledError:
            raise RuntimeError("Coroutine was cancelled")
        except Exception as e:
            self._log.error(f"Error running coroutine: {e}")
            raise

    def cancel_task(self, name: str) -> bool:
        if name in self._tasks:
            self._tasks[name].cancel()
            return True
        return False

    def _handle_task_done(self, task: asyncio.Task):
        try:
            name = task.get_name()
            self._tasks.pop(name, None)
            task.result()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self._log.error(f"Error during task done: {e}")
            raise

    async def wait(self):
        try:
            if self._tasks:
                await self._shutdown_event.wait()
        except Exception as e:
            self._log.error(f"Error during wait: {e}")
            raise

    async def cancel(self, timeout: float | None = None):
        try:
            # Snapshot tasks to avoid dict mutation during iteration
            tasks = [t for t in self._tasks.values() if not t.done()]

            for task in tasks:
                task.cancel()

            if tasks:
                gather_coro = asyncio.gather(*tasks, return_exceptions=True)
                try:
                    effective_timeout = (
                        self._cancel_timeout if timeout is None else timeout
                    )
                    if effective_timeout is not None:
                        results = await asyncio.wait_for(
                            gather_coro, timeout=effective_timeout
                        )
                    else:
                        results = await gather_coro
                except asyncio.TimeoutError:
                    self._log.warning(
                        "Cancellation timed out; some tasks may still be running"
                    )
                else:
                    for result in results:
                        if isinstance(result, Exception) and not isinstance(
                            result, asyncio.CancelledError
                        ):
                            self._log.error(
                                f"Task failed during cancellation: {result}"
                            )

        except Exception as e:
            self._log.error(f"Error during cancellation: {e}")
            raise
        finally:
            self._tasks.clear()


# class Clock:
#     def __init__(self, tick_size: float = 1.0):
#         """
#         :param tick_size_s: Time interval of each tick in seconds (supports sub-second precision).
#         """
#         self._tick_size = tick_size  # Tick size in seconds
#         self._current_tick = (time.time() // self._tick_size) * self._tick_size
#         self._clock = LiveClock()
#         self._tick_callbacks: List[Callable[[float], None]] = []
#         self._started = False

#     @property
#     def tick_size(self) -> float:
#         return self._tick_size

#     @property
#     def current_timestamp(self) -> float:
#         return self._clock.timestamp()

#     def add_tick_callback(self, callback: Callable[[float], None]):
#         """
#         Register a callback to be called on each tick.
#         :param callback: Function to be called with current_tick as argument.
#         """
#         self._tick_callbacks.append(callback)

#     async def run(self):
#         if self._started:
#             raise RuntimeError("Clock is already running.")
#         self._started = True
#         while True:
#             now = time.time()
#             next_tick_time = self._current_tick + self._tick_size
#             sleep_duration = next_tick_time - now
#             if sleep_duration > 0:
#                 await asyncio.sleep(sleep_duration)
#             else:
#                 # If we're behind schedule, skip to the next tick to prevent drift
#                 next_tick_time = now
#             self._current_tick = next_tick_time
#             for callback in self._tick_callbacks:
#                 if asyncio.iscoroutinefunction(callback):
#                     await callback(self.current_timestamp)
#                 else:
#                     callback(self.current_timestamp)


class ZeroMQSignalRecv:
    def __init__(self, config, callback: Callable, task_manager: TaskManager):
        self._socket = config.socket
        self._callback = callback
        self._task_manager = task_manager

    async def _recv(self):
        while True:
            date = await self._socket.recv()
            if asyncio.iscoroutinefunction(self._callback):
                await self._callback(date)
            else:
                self._callback(date)

    async def start(self):
        self._task_manager.create_task(self._recv())


# class MovingAverage:
#     """
#     Calculate moving median or mean using a sliding window.

#     Args:
#         length: Length of the sliding window
#         method: 'median' or 'mean' calculation method
#     """

#     def __init__(self, length: int, method: Literal["median", "mean"] = "mean"):
#         if method not in ["median", "mean"]:
#             raise ValueError("method must be either 'median' or 'mean'")

#         self._length = length
#         self._method = method
#         self._window = deque(maxlen=length)
#         self._calc_func = median if method == "median" else mean

#     def input(self, value: float) -> float | None:
#         """
#         Input a new value and return the current median/mean.

#         Args:
#             value: New value to add to sliding window

#         Returns:
#             Current median/mean value, or None if window not filled
#         """
#         self._window.append(value)

#         if len(self._window) < self._length:
#             return None

#         return self._calc_func(self._window)


class DataReady:
    def __init__(
        self,
        symbols: List[str],
        name: str,
        clock: LiveClock,
        timeout: int = 60,
        permanently_ready: bool = False,
    ):
        """
        Initialize DataReady class

        Args:
            symbols: symbols list need to monitor
            timeout: timeout in seconds
        """
        self._log = logging.getLogger(name=type(self).__name__)

        # Optimization: Store symbols as a set for faster lookups if needed,
        # but dict is fine for tracking status.
        self._symbols_status = {symbol: False for symbol in symbols}
        self._total_symbols = len(symbols)
        self._ready_symbols_count = 0

        self._timeout_ms = timeout * 1000  # Store timeout in ms
        self._clock = clock
        self._first_data_time: int | None = None
        self._name = name
        # Optimization: A flag to indicate that the "ready" state is final
        # (either all symbols received or timed out).
        self._is_permanently_ready: bool = permanently_ready

        if not symbols:  # If no symbols to monitor, it's ready immediately (or upon first data for timestamp)
            # We'll consider it ready once _first_data_time is set, which requires at least one input call
            # Or, we can decide it's ready right away if that's the desired behavior.
            # For now, let's assume if symbols is empty, it becomes ready when `ready` is first checked
            # after _first_data_time is set, or on timeout.
            # Alternatively, one could set self._is_permanently_ready = True here.
            # The current logic for _ready_symbols_count == _total_symbols (0 == 0) will handle this.
            self._is_permanently_ready = True

    def input(self, data: InputDataType) -> None:
        """
        Input data, update the status of the corresponding symbol

        Args:
            data: data object with symbol attribute
        """
        # Optimization: If already permanently ready, do nothing.
        if self._is_permanently_ready:
            return

        symbol = data.symbol

        if self._first_data_time is None:
            self._first_data_time = self._clock.timestamp_ms()

        # Process only if the symbol is one we are monitoring and it's not yet marked ready.
        if symbol in self._symbols_status and not self._symbols_status[symbol]:
            self._symbols_status[symbol] = True
            self._ready_symbols_count += 1

            # Optimization: Check if all symbols are now ready.
            if self._ready_symbols_count == self._total_symbols:
                self._log.debug(
                    f"All {self._total_symbols} symbols received. {self._name} is ready."
                )
                self._is_permanently_ready = True
                # No need to call self.ready here, the flag is enough.

    def add_symbols(self, symbols: str | list[str]) -> None:
        """
        Add a new symbol to monitor

        Args:
            symbol: symbol to add
        """
        if isinstance(symbols, str):
            symbols = [symbols]

        for symbol in symbols:
            if symbol not in self._symbols_status:
                self._symbols_status[symbol] = False
                self._total_symbols += 1
                # If we were already permanently ready, adding a new symbol means we are no longer ready.
                if self._is_permanently_ready:
                    self._is_permanently_ready = False

    @property
    def ready(self) -> bool:
        """
        Check if all data is ready or if it has timed out

        Returns:
            bool: if all data is ready or timed out, return True
        """
        # Optimization: If already determined to be permanently ready, return True.
        if self._is_permanently_ready:
            return True

        if self._first_data_time is None:
            # No data received yet, so not ready (unless symbols list was empty and we decided that's ready from init)
            return False  # Or handle empty symbols list case from __init__

        # Check if all symbols have been received (could have been set by input)
        if self._ready_symbols_count == self._total_symbols:
            # This ensures that even if input didn't set _is_permanently_ready
            # (e.g. if ready is called before next input after last symbol arrived),
            # it's correctly identified.
            if (
                not self._is_permanently_ready
            ):  # To avoid repeated logging if already logged by input
                self._log.debug(
                    f"All {self._total_symbols} symbols confirmed ready by property access."
                )
            self._is_permanently_ready = True
            return True

        # Check for timeout
        if self._clock.timestamp_ms() - self._first_data_time > self._timeout_ms:
            # Only issue warning if not all symbols were ready before timeout
            if self._ready_symbols_count < self._total_symbols:
                not_ready_symbols = [
                    symbol
                    for symbol, status in self._symbols_status.items()
                    if not status
                ]
                if not_ready_symbols:  # Should always be true if count < total
                    warnings.warn(
                        f"Data receiving timed out. Timeout: {self._timeout_ms / 1000}s. "
                        f"Received {self._ready_symbols_count}/{self._total_symbols} symbols. "
                        f"The following symbols are not ready: {', '.join(not_ready_symbols)}"
                    )
            else:  # All symbols were received, but timeout check ran.
                self._log.debug(
                    f"Timeout checked, but all symbols were already received. {self._name} is ready."
                )

            self._is_permanently_ready = True  # Timed out, so state is now final.
            return True

        return False  # Not all symbols ready and not timed out yet.



"""
Kline aggregation for live trading.

This module provides kline (candlestick) aggregation from trade data.
"""

from typing import Optional, Literal
import nexuslog as logging
from datetime import timedelta, datetime, timezone
from nexustrader.schema import Trade, Kline
from nexustrader.constants import KlineInterval, ExchangeType, OrderSide
from nexustrader.core.nautilius_core import LiveClock, TimeEvent, MessageBus


class KlineBuilder:
    """
    Kline builder for aggregating trade data into OHLCV klines.

    Parameters
    ----------
    exchange : ExchangeType
        The exchange for the klines
    symbol : str
        The symbol for the klines
    interval : KlineInterval
        The kline interval
    """

    def __init__(
        self,
        exchange: ExchangeType,
        symbol: str,
        interval: KlineInterval,
    ):
        self.exchange = exchange
        self.symbol = symbol
        self.interval = interval

        # OHLCV data as member variables
        self._open = None
        self._high = None
        self._low = None
        self._close = None
        self._volume = 0.0
        self._buy_volume = 0.0
        self._last_close = None

        self.initialized = False
        self.ts_last = 0
        self.count = 0

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"{self.exchange.value}, "
            f"{self.symbol}, "
            f"{self.interval.value}, "
            f"open={self._open}, "
            f"high={self._high}, "
            f"low={self._low}, "
            f"close={self._close}, "
            f"volume={self._volume}, "
            f"buy_volume={self._buy_volume})"
        )

    def update(self, trade: Trade) -> None:
        """
        Update the kline builder with a new trade.

        Parameters
        ----------
        trade : Trade
            The trade data to update with
        """
        if trade.timestamp < self.ts_last:
            return

        if trade.price <= 0:
            return

        if self._open is None:
            self._open = trade.price
            self._high = trade.price
            self._low = trade.price
            self.initialized = True
        elif trade.price > self._high:
            self._high = trade.price
        elif trade.price < self._low:
            self._low = trade.price

        self._close = trade.price
        self._volume += trade.size
        if trade.side.is_buy:
            self._buy_volume += trade.size
        self.count += 1
        self.ts_last = trade.timestamp

    def reset(self) -> None:
        """Reset the builder to initial state."""
        self._open = None
        self._high = None
        self._low = None
        self._volume = 0.0
        self._buy_volume = 0.0
        self.count = 0

    def build(self, start: int, timestamp: int) -> Optional[Kline]:
        """
        Build a kline from the current state and reset.

        Returns None if no trades were received (not initialized).

        Parameters
        ----------
        ts_event : int
            Timestamp (nanoseconds) for the kline event
        ts_init : int
            Timestamp (nanoseconds) for the kline initialization

        Returns
        -------
        Kline | None
            The built kline, or None if no trades received
        """
        # If no trades received, don't emit a kline
        if not self.initialized:
            return None

        if self._open is None:
            self._open = self._last_close
            self._high = self._last_close
            self._low = self._last_close
            self._close = self._last_close

        self._low = min(self._low, self._close)
        self._high = max(self._high, self._close)

        kline = Kline(
            exchange=self.exchange,
            symbol=self.symbol,
            interval=self.interval,
            open=self._open,
            high=self._high,
            low=self._low,
            close=self._close,
            volume=self._volume,
            buy_volume=self._buy_volume,
            start=start,
            timestamp=timestamp,
            confirm=True,
        )

        self._last_close = self._close
        self.reset()
        return kline

    def build_now(self) -> Optional[Kline]:
        """
        Build a kline with current timestamp and reset.

        Returns None if no trades were received.

        Returns
        -------
        Kline | None
            The built kline, or None if no trades received
        """
        return self.build(self.ts_last, self.ts_last)


class KlineAggregator:
    """
    Base class for kline aggregation from trade data.

    Parameters
    ----------
    exchange : ExchangeType
        The exchange for the aggregator
    symbol : str
        The symbol for the aggregator
    interval : KlineInterval
        The kline interval
    msgbus : MessageBus
        The message bus for publishing klines
    """

    def __init__(
        self,
        exchange: ExchangeType,
        symbol: str,
        msgbus: MessageBus,
        interval: KlineInterval,
    ):
        self.exchange = exchange
        self.symbol = symbol
        self.interval = interval
        self._msgbus = msgbus
        self._log = logging.getLogger(name=type(self).__name__)

        self._builder = KlineBuilder(
            exchange=exchange,
            symbol=symbol,
            interval=interval,
        )
        self.is_running = False

    def handle_trade(self, trade: Trade) -> None:
        """
        Handle incoming trade data.

        Parameters
        ----------
        trade : Trade
            The trade to process
        """
        self._apply_update(trade)

    def _apply_update(self, trade: Trade) -> None:
        """
        Apply trade update to the aggregator.

        Must be implemented by subclasses.

        Parameters
        ----------
        trade : Trade
            The trade to process
        """
        raise NotImplementedError(
            "method `_apply_update` must be implemented in the subclass"
        )

    def _build_now_and_send(self) -> None:
        """Build kline with current timestamp and publish to msgbus if kline was built."""
        kline = self._builder.build_now()
        if kline is not None:
            self._msgbus.publish(topic="kline", msg=kline)

    def _build_and_send(self, start: int, timestamp: int) -> None:
        """Build kline with specified timestamps and publish to msgbus if kline was built."""
        kline = self._builder.build(start, timestamp)
        if kline is not None:
            self._msgbus.publish(topic="kline", msg=kline)


class VolumeKlineAggregator(KlineAggregator):
    """
    Volume-based kline aggregator.

    Creates klines when cumulative trade volume reaches the threshold.
    Large trades are split across multiple klines.

    Parameters
    ----------
    exchange : ExchangeType
        The exchange for the aggregator
    symbol : str
        The symbol for the aggregator
    interval : KlineInterval
        The kline interval (used for metadata only)
    msgbus : MessageBus
        The message bus for publishing klines
    volume_threshold : float
        Volume threshold for creating new klines
    """

    def __init__(
        self,
        exchange: ExchangeType,
        symbol: str,
        msgbus: MessageBus,
        volume_threshold: float,
        volume_type: Literal["DEFAULT", "BUY", "SELL"] = "DEFAULT",
    ):
        super().__init__(exchange, symbol, msgbus, interval=KlineInterval.VOLUME)
        self.volume_threshold = volume_threshold

        if volume_type == "DEFAULT":
            self._side = None
        elif volume_type == "BUY":
            self._side = OrderSide.BUY
        elif volume_type == "SELL":
            self._side = OrderSide.SELL
        else:
            raise ValueError(f"Invalid volume_type: {volume_type}")

    def _apply_update(self, trade: Trade) -> None:
        """
        Apply trade update with volume-based aggregation.

        Parameters
        ----------
        trade : Trade
            The trade to process
        """
        if self._side is not None and trade.side != self._side:
            return

        size_update = trade.size

        while size_update > 0:
            current_volume = self._builder._volume

            if current_volume + size_update < self.volume_threshold:
                # Update and break
                partial_trade = Trade(
                    exchange=trade.exchange,
                    symbol=trade.symbol,
                    price=trade.price,
                    size=size_update,
                    side=trade.side,
                    timestamp=trade.timestamp,
                )
                self._builder.update(partial_trade)
                break

            # Calculate size needed to reach threshold
            size_diff = self.volume_threshold - current_volume

            # Update builder to threshold
            partial_trade = Trade(
                exchange=trade.exchange,
                symbol=trade.symbol,
                side=trade.side,
                price=trade.price,
                size=size_diff,
                timestamp=trade.timestamp,
            )
            self._builder.update(partial_trade)
            self._build_now_and_send()

            # Decrement remaining size
            size_update -= size_diff


class TimeKlineAggregator(KlineAggregator):
    """
    Time-based kline aggregator using LiveClock.

    Creates klines at regular time intervals (1s to 1w).

    Parameters
    ----------
    exchange : ExchangeType
        The exchange for the aggregator
    symbol : str
        The symbol for the aggregator
    interval : KlineInterval
        The kline interval
    msgbus : MessageBus
        The message bus for publishing klines
    clock : LiveClock
        The clock for timing
    """

    def __init__(
        self,
        exchange: ExchangeType,
        symbol: str,
        interval: KlineInterval,
        msgbus: MessageBus,
        clock: LiveClock,
        build_with_no_updates: bool = True,
    ):
        super().__init__(exchange, symbol, msgbus, interval)
        self._clock = clock
        self._timer_name = f"{exchange.value}_{symbol}_{interval.value}"

        self._interval_ms = interval.milliseconds
        self._build_with_no_updates = build_with_no_updates

        # Set up the timer
        self._set_build_timer()

    def _set_build_timer(self) -> None:
        """Set up the clock timer for kline building."""
        start_time = self._get_start_time()

        # Calculate interval timedelta
        interval_td = timedelta(milliseconds=self._interval_ms)

        self._clock.set_timer(
            name=self._timer_name,
            interval=interval_td,
            start_time=start_time,
            stop_time=None,
            callback=self._build_bar,
        )
        self._log.debug(
            f"Timer set: {self._timer_name}, start={start_time}, interval={interval_td}"
        )

    def _get_start_time(self):
        """
        Calculate the start time for the next kline interval.

        Uses the same logic as Nautilus TimeBarAggregator:
        - Find the floor time (closest smaller time aligned to interval)
        - Always schedule for the NEXT interval boundary
        """
        timestamp = self._clock.timestamp_ms()
        interval_ms = self._interval_ms
        floored_timestamp = (timestamp // interval_ms) * interval_ms
        start_time = datetime.fromtimestamp(floored_timestamp / 1000, tz=timezone.utc)

        # Always start at the NEXT interval boundary
        return start_time + timedelta(milliseconds=interval_ms)

    def stop(self) -> None:
        """Stop the aggregator and cancel the timer."""
        self._clock.cancel_timer(self._timer_name)

    def _apply_update(self, trade: Trade) -> None:
        """
        Apply trade update with time-based aggregation.

        Parameters
        ----------
        trade : Trade
            The trade to process
        """
        self._builder.update(trade)

    def _build_bar(self, event: TimeEvent) -> None:
        """
        Build and emit kline on timer event.

        Parameters
        ----------
        event : TimeEvent
            The timer event
        """
        if not self._build_with_no_updates and self._builder.count == 0:
            return
        ts_event = event.ts_event // 1_000_000 - self._interval_ms  # Convert ns to ms
        ts_init = event.ts_init // 1_000_000  # Convert ns to ms

        # # Build and send kline (only if trades were received)
        self._build_and_send(ts_event, ts_init)
        self._log.debug(f"Kline built: ts_event={ts_event}, ts_init={ts_init}")



import asyncio
from decimal import Decimal
from typing import Dict, List
from nexustrader.constants import AccountType, SubmitType
from nexustrader.schema import (
    CreateOrderSubmit,
    OrderSubmit,
    InstrumentId,
    BatchOrderSubmit,
)
from nexustrader.core.cache import AsyncCache
from nexustrader.core.nautilius_core import MessageBus, LiveClock
from nexustrader.core.entity import TaskManager
from nexustrader.core.registry import OrderRegistry
from nexustrader.exchange.hyperliquid import HyperLiquidAccountType
from nexustrader.exchange.hyperliquid.schema import HyperLiquidMarket
from nexustrader.exchange.hyperliquid.constants import oid_to_cloid_hex
from nexustrader.base import ExecutionManagementSystem
from nexustrader.schema import CancelAllOrderSubmit, CancelOrderSubmit


class HyperLiquidExecutionManagementSystem(ExecutionManagementSystem):
    _market: Dict[str, HyperLiquidMarket]

    HYPER_LIQUID_ACCOUNT_TYPE_PRIORITY = [
        HyperLiquidAccountType.MAINNET,
        HyperLiquidAccountType.TESTNET,
    ]

    def __init__(
        self,
        market: Dict[str, HyperLiquidMarket],
        cache: AsyncCache,
        msgbus: MessageBus,
        clock: LiveClock,
        task_manager: TaskManager,
        registry: OrderRegistry,
        is_mock: bool = False,
    ):
        super().__init__(
            market=market,
            cache=cache,
            msgbus=msgbus,
            clock=clock,
            task_manager=task_manager,
            registry=registry,
            is_mock=is_mock,
        )
        self._hyperliquid_account_type: HyperLiquidAccountType = None

    def _build_order_submit_queues(self):
        for account_type in self._private_connectors.keys():
            if isinstance(account_type, HyperLiquidAccountType):
                self._order_submit_queues[account_type] = asyncio.Queue()
                break

    def _set_account_type(self):
        account_types = self._private_connectors.keys()
        for account_type in self.HYPER_LIQUID_ACCOUNT_TYPE_PRIORITY:
            if account_type in account_types:
                self._hyperliquid_account_type = account_type
                break

    def _instrument_id_to_account_type(
        self, instrument_id: InstrumentId
    ) -> AccountType:
        if self._is_mock:
            if instrument_id.is_spot:
                return HyperLiquidAccountType.SPOT_MOCK
            elif instrument_id.is_linear:
                return HyperLiquidAccountType.LINEAR_MOCK
            elif instrument_id.is_inverse:
                return HyperLiquidAccountType.INVERSE_MOCK
        else:
            return self._hyperliquid_account_type

    def _submit_order(
        self,
        order: OrderSubmit | List[OrderSubmit],
        submit_type: SubmitType,
        account_type: AccountType | None = None,
    ):
        if isinstance(order, list):
            if not account_type:
                account_type = self._instrument_id_to_account_type(
                    order[0].instrument_id
                )

            # Split batch orders into chunks of 20
            for i in range(0, len(order), 20):
                batch = order[i : i + 20]
                self._order_submit_queues[account_type].put_nowait((batch, submit_type))
        else:
            if not account_type:
                account_type = self._instrument_id_to_account_type(order.instrument_id)
            self._order_submit_queues[account_type].put_nowait((order, submit_type))

    def _get_min_order_amount(
        self, symbol: str, market: HyperLiquidMarket, px: float
    ) -> Decimal:
        # book = self._cache.bookl1(symbol)
        min_order_cost = market.limits.cost.min
        min_order_amount = super()._amount_to_precision(
            symbol, min_order_cost / px * 1.01, mode="ceil"
        )
        return min_order_amount

    async def _cancel_all_orders(
        self, order_submit: CancelAllOrderSubmit, account_type: AccountType
    ):
        # override the base method
        symbol = order_submit.symbol
        oids = self._cache.get_open_orders(symbol)
        for oid in oids:
            order_submit = CancelOrderSubmit(
                symbol=symbol,
                instrument_id=InstrumentId.from_str(symbol),
                oid=oid,
            )
            await self._cancel_order(order_submit, account_type)

    async def _create_order(
        self, order_submit: CreateOrderSubmit, account_type: AccountType
    ):
        """
        Create an order
        """
        oid = oid_to_cloid_hex(order_submit.oid)
        self._registry.register_order(oid)
        await self._private_connectors[account_type]._oms.create_order(
            oid=oid,
            symbol=order_submit.symbol,
            side=order_submit.side,
            type=order_submit.type,
            amount=order_submit.amount,
            price=order_submit.price,
            time_in_force=order_submit.time_in_force,
            reduce_only=order_submit.reduce_only,
            **order_submit.kwargs,
        )

    async def _create_batch_orders(
        self, batch_orders: List[BatchOrderSubmit], account_type: AccountType
    ):
        new_batch_orders = []
        for order in batch_orders:
            oid = oid_to_cloid_hex(order.oid)
            self._registry.register_order(oid=oid)
            new_batch_orders.append(
                BatchOrderSubmit(
                    symbol=order.symbol,
                    instrument_id=order.instrument_id,
                    status=order.status,
                    side=order.side,
                    type=order.type,
                    amount=order.amount,
                    price=order.price,
                    time_in_force=order.time_in_force,
                    reduce_only=order.reduce_only,
                    oid=oid,
                    kwargs=order.kwargs,
                )
            )
        await self._private_connectors[account_type]._oms.create_batch_orders(
            orders=new_batch_orders,
        )

    async def _create_order_ws(
        self, order_submit: CreateOrderSubmit, account_type: AccountType
    ):
        """
        Create an order
        """
        oid = oid_to_cloid_hex(order_submit.oid)
        self._registry.register_order(oid)
        await self._private_connectors[account_type]._oms.create_order_ws(
            oid=oid,
            symbol=order_submit.symbol,
            side=order_submit.side,
            type=order_submit.type,
            amount=order_submit.amount,
            price=order_submit.price,
            time_in_force=order_submit.time_in_force,
            reduce_only=order_submit.reduce_only,
            **order_submit.kwargs,
        )



"""Strategy-aware FastAPI app utilities."""

from __future__ import annotations

import inspect
from dataclasses import dataclass
from functools import update_wrapper
from typing import Any, Callable, Dict, List, Optional

from fastapi import FastAPI
from fastapi.routing import APIRouter, APIRoute


@dataclass
class _StrategyRoute:
    path: str
    endpoint: Callable[..., Any]
    kwargs: Dict[str, Any]
    bound_route: Optional[APIRoute] = None
    bound_endpoint: Optional[Callable[..., Any]] = None


class StrategyAPIRouter(APIRouter):
    """APIRouter that defers binding of strategy instance methods."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._strategy_routes: List[_StrategyRoute] = []
        self._baseline_route_count = len(self.routes)

    # when @app.post(...) is called, this method is invoked
    def add_api_route(
        self, path: str, endpoint: Callable[..., Any], **kwargs: Any
    ) -> None:  # type: ignore[override]
        if self._is_strategy_endpoint(
            endpoint
        ):  # first parameter is 'self', indicating a method, save it to self._strategy_routes
            self._strategy_routes.append(
                _StrategyRoute(path=path, endpoint=endpoint, kwargs=dict(kwargs))
            )
            return None
        return super().add_api_route(path, endpoint, **kwargs)

    def bind_strategy(self, strategy: Any) -> None:
        for route_def in self._strategy_routes:
            self._remove_existing_route(route_def)
            bound = self._wrap_endpoint(route_def.endpoint, strategy)
            super().add_api_route(route_def.path, bound, **route_def.kwargs)
            # Newly added route is the last item in self.routes
            route_def.bound_route = self.routes[-1]  # type: ignore[assignment]
            route_def.bound_endpoint = bound

    def unbind_strategy(self) -> None:
        for route_def in self._strategy_routes:
            if route_def.bound_route and route_def.bound_route in self.routes:
                self.routes.remove(route_def.bound_route)
                route_def.bound_route = None
                route_def.bound_endpoint = None

    def has_strategy_routes(self) -> bool:
        return bool(self._strategy_routes)

    def has_user_defined_routes(self) -> bool:
        return (
            self.has_strategy_routes() or len(self.routes) > self._baseline_route_count
        )

    @staticmethod
    def _is_strategy_endpoint(endpoint: Callable[..., Any]) -> bool:
        try:
            signature = inspect.signature(endpoint)
        except (TypeError, ValueError):
            return False

        params = list(signature.parameters.values())
        if not params:
            return False
        return params[0].name == "self"

    @staticmethod
    def _wrap_endpoint(
        endpoint: Callable[..., Any], strategy: Any
    ) -> Callable[..., Any]:
        signature = inspect.signature(endpoint)
        params = list(signature.parameters.values())
        if not params or params[0].name != "self":
            return endpoint

        stripped_signature = signature.replace(parameters=params[1:])

        if inspect.iscoroutinefunction(endpoint):

            async def _call(*args: Any, **kwargs: Any) -> Any:
                return await endpoint(strategy, *args, **kwargs)

        else:

            def _call(*args: Any, **kwargs: Any) -> Any:
                return endpoint(strategy, *args, **kwargs)

        update_wrapper(_call, endpoint)
        _call.__signature__ = stripped_signature  # type: ignore[attr-defined]
        return _call

    def _remove_existing_route(self, route_def: _StrategyRoute) -> None:
        if not route_def.bound_route:
            return
        if route_def.bound_route in self.routes:
            self.routes.remove(route_def.bound_route)
        route_def.bound_route = None
        route_def.bound_endpoint = None


class StrategyFastAPI(FastAPI):
    """FastAPI application that understands strategy instance methods."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        original_router: APIRouter = self.router
        router = StrategyAPIRouter(
            prefix=original_router.prefix,
            tags=original_router.tags,
            dependencies=original_router.dependencies,
            default_response_class=original_router.default_response_class,
            responses=original_router.responses,
            callbacks=original_router.callbacks,
            routes=list(original_router.routes),
            deprecated=original_router.deprecated,
            include_in_schema=original_router.include_in_schema,
            generate_unique_id_function=original_router.generate_unique_id_function,
        )
        router.route_class = original_router.route_class
        router.on_startup = original_router.on_startup
        router.on_shutdown = original_router.on_shutdown
        router.lifespan_context = original_router.lifespan_context

        self.router = router
        self._strategy_router: StrategyAPIRouter = router

    def bind_strategy(self, strategy: Any) -> None:
        self._strategy_router.bind_strategy(strategy)

    def unbind_strategy(self) -> None:
        self._strategy_router.unbind_strategy()

    def has_strategy_routes(self) -> bool:
        return self._strategy_router.has_strategy_routes()

    def has_user_routes(self) -> bool:
        return self._strategy_router.has_user_defined_routes()


def create_strategy_app(**kwargs: Any) -> StrategyFastAPI:
    """Create a strategy-aware FastAPI application."""

    return StrategyFastAPI(**kwargs)



import nexuslog as logging
from nexustrader.constants import settings


use_nautilius = settings.get("USE_NAUTILIUS", False)

if use_nautilius:
    try:
        from nautilus_trader.common.component import MessageBus
        from nautilus_trader.common.component import LiveClock, TimeEvent
        from nautilus_trader.model.identifiers import TraderId
        from nautilus_trader.core.uuid import UUID4

        from nautilus_trader.core import nautilus_pyo3  # noqa
        from nautilus_trader.core.nautilus_pyo3 import HttpClient  # noqa
        from nautilus_trader.core.nautilus_pyo3 import HttpMethod  # noqa
        from nautilus_trader.core.nautilus_pyo3 import HttpResponse  # noqa

        # from nautilus_trader.core.nautilus_pyo3 import MessageBus  # noqa
        from nautilus_trader.core.nautilus_pyo3 import WebSocketClient  # noqa
        from nautilus_trader.core.nautilus_pyo3 import WebSocketClientError  # noqa
        from nautilus_trader.core.nautilus_pyo3 import WebSocketConfig  # noqa
        from nautilus_trader.core.nautilus_pyo3 import (
            hmac_signature,  # noqa
            rsa_signature,  # noqa
            ed25519_signature,  # noqa
        )
    except ImportError as e:
        raise ImportError(
            "Nautilus Trader is not installed. Please install run `pip install nautilus-trader` to use Nautilus features."
        ) from e
else:
    from nexuscore import (
        MessageBus,
        LiveClock,
        TimeEvent,
        TraderId,
        UUID4,
        hmac_signature,
        rsa_signature,
        ed25519_signature,
    )


def setup_nexus_core(
    trader_id: str,
    filename: str | None = None,
    level: str = "INFO",
    name_levels: dict[str | None, str] | None = None,
    unix_ts: bool = False,
    batch_size: int | None = None,
):
    """
    Setup logging for the application using nexuslog and initialize MessageBus and Clock.

    Args:
        trader_id: Unique identifier for the trader
        filename: Optional file path for log output. If None, logs to stdout.
        level: Minimum log level to record (TRACE, DEBUG, INFO, WARNING, ERROR)
        unix_ts: If True, emit unix timestamps instead of formatted local time.

    Returns:
        tuple: (msgbus, clock) - MessageBus and LiveClock instances
    """
    clock = LiveClock()
    msgbus = MessageBus(
        trader_id=TraderId(trader_id),
        clock=clock,
    )

    # Map log levels from string to nexuslog levels
    level_map = {
        "TRACE": logging.TRACE,
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
    }

    log_level = level_map.get(level, logging.INFO)

    _name_levels = None
    if name_levels:
        _name_levels = {
            name: level_map.get(lvl_str, logging.INFO)
            for name, lvl_str in name_levels.items()
        }

    # Configure nexuslog using basicConfig
    logging.basicConfig(
        filename=filename,
        level=log_level,
        name_levels=_name_levels,
        unix_ts=unix_ts,
        batch_size=batch_size,
    )

    return msgbus, clock



from dataclasses import dataclass
from typing import Literal


ConnectionRole = Literal["MD", "TD"]


@dataclass(frozen=True)
class ConnectionState:
    role: ConnectionRole
    exchange_id: str
    account_type: str
    ws_name: str
    client_id: int
    required: bool
    connected: bool
    changed_at_ms: int

    def is_md(self) -> bool:
        return self.role == "MD"

    def is_td(self) -> bool:
        return self.role == "TD"


@dataclass(frozen=True)
class ConnectionPolicyState:
    md_ok: bool
    td_ok: bool

    @property
    def allow_open(self) -> bool:
        return self.md_ok and self.td_ok

    @property
    def allow_trade(self) -> bool:
        return self.td_ok

    @property
    def allow_close_only(self) -> bool:
        return self.td_ok and not self.md_ok



from typing import Any


class HyperLiquidHttpError(Exception):
    def __init__(self, status_code: int, message: str, headers: dict[str, Any]):
        super().__init__(message)
        self.status_code = status_code
        self.message = message
        self.headers = headers

    def __repr__(self) -> str:
        return f"{type(self).__name__}(status_code={self.status_code}, message='{self.message}')"

    __str__ = __repr__


class HyperLiquidOrderError(Exception):
    """
    The base class for all HyperLiquid specific errors.
    """

    def __init__(self, error_type: str, message: str):
        super().__init__(message)
        self.error_type = error_type
        self.message = message

    def __repr__(self) -> str:
        return f"{type(self).__name__}(error_type={self.error_type}, message='{self.message}')"

    __str__ = __repr__



import msgspec
from decimal import Decimal
from typing import Dict, List
from nexustrader.error import PositionModeError
from nexustrader.constants import (
    ExchangeType,
    OrderSide,
    OrderType,
    TimeInForce,
    OrderStatus,
    TriggerType,
    PositionSide,
)
from nexustrader.exchange.hyperliquid.rest_api import HyperLiquidApiClient
from nexustrader.base import OrderManagementSystem
from nexustrader.core.registry import OrderRegistry
from nexustrader.core.nautilius_core import LiveClock, MessageBus
from nexustrader.core.entity import TaskManager
from nexustrader.core.cache import AsyncCache
from nexustrader.schema import (
    Balance,
    Order,
    BatchOrderSubmit,
    Position,
)
from nexustrader.exchange.hyperliquid.schema import (
    HyperLiquidMarket,
    HyperLiquidWsMessageGeneral,
    HyperLiquidWsOrderUpdatesMsg,
    HyperLiquidWsUserFillsMsg,
    HyperLiquidWsApiGeneralMsg,
    HyperLiquidWsApiOrderData,
    HyperLiquidWsApiMessageData,
    HyperLiquidWsApiOrderStatus,
)
from nexustrader.exchange.hyperliquid.websockets import (
    HyperLiquidWSClient,
    HyperLiquidWSApiClient,
)
from nexustrader.exchange.hyperliquid.constants import (
    HyperLiquidAccountType,
    HyperLiquidEnumParser,
    HyperLiquidOrderRequest,
    HyperLiquidCloidCancelRequest,
    HyperLiquidTimeInForce,
)


class HyperLiquidOrderManagementSystem(OrderManagementSystem):
    _ws_client: HyperLiquidWSClient
    _ws_api_client: HyperLiquidWSApiClient
    _account_type: HyperLiquidAccountType
    _market: Dict[str, HyperLiquidMarket]
    _market_id: Dict[str, str]
    _api_client: HyperLiquidApiClient

    def __init__(
        self,
        account_type: HyperLiquidAccountType,
        api_key: str,
        secret: str,
        market: Dict[str, HyperLiquidMarket],
        market_id: Dict[str, str],
        registry: OrderRegistry,
        cache: AsyncCache,
        api_client: HyperLiquidApiClient,
        exchange_id: ExchangeType,
        clock: LiveClock,
        msgbus: MessageBus,
        task_manager: TaskManager,
        max_slippage: float,
        max_subscriptions_per_client: int | None = None,
        max_clients: int | None = None,
    ):
        super().__init__(
            account_type=account_type,
            market=market,
            market_id=market_id,
            registry=registry,
            cache=cache,
            api_client=api_client,
            ws_client=HyperLiquidWSClient(
                account_type=account_type,
                handler=self._ws_msg_handler,
                clock=clock,
                task_manager=task_manager,
                api_key=api_key,
                max_subscriptions_per_client=max_subscriptions_per_client,
                max_clients=max_clients,
            ),
            exchange_id=exchange_id,
            clock=clock,
            msgbus=msgbus,
        )

        self._max_slippage = max_slippage
        self._ws_msg_general_decoder = msgspec.json.Decoder(HyperLiquidWsMessageGeneral)
        self._ws_msg_order_updates_decoder = msgspec.json.Decoder(
            HyperLiquidWsOrderUpdatesMsg
        )
        self._ws_msg_user_events_decoder = msgspec.json.Decoder(
            HyperLiquidWsUserFillsMsg
        )

        # Initialize WebSocket API client
        self._ws_api_client = HyperLiquidWSApiClient(
            account_type=account_type,
            api_key=api_key,
            secret=secret,
            handler=self._ws_api_msg_handler,
            task_manager=task_manager,
            clock=clock,
            enable_rate_limit=True,
        )
        self._ws_api_msg_decoder = msgspec.json.Decoder(HyperLiquidWsApiGeneralMsg)

    def _init_account_balance(self):
        """Initialize the account balance"""
        res = self._api_client.get_user_spot_summary()
        self._cache._apply_balance(
            account_type=self._account_type, balances=res.parse_to_balances()
        )

    def _init_position(self):
        """Initialize the position"""
        res = self._api_client.get_user_perps_summary()
        for pos_data in res.assetPositions:
            if pos_data.type != "oneWay":
                raise PositionModeError(
                    f"HyperLiquid only supports one-way position mode, but got {pos_data.type}"
                )

            symbol = self._market_id.get(pos_data.position.coin, None)
            if not symbol:
                continue

            signed_amount = Decimal(pos_data.position.szi)
            if signed_amount == Decimal("0"):
                continue

            position = Position(
                symbol=symbol,
                exchange=self._exchange_id,
                signed_amount=Decimal(pos_data.position.szi),
                side=PositionSide.LONG
                if signed_amount > Decimal("0")
                else PositionSide.SHORT,
                entry_price=float(pos_data.position.entryPx),
                unrealized_pnl=float(pos_data.position.unrealizedPnl),
            )
            self._cache._apply_position(position)

    def _position_mode_check(self):
        """Check the position mode"""
        # NOTE: HyperLiquid only supports one-way position mode
        pass

    async def create_tp_sl_order(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal | None = None,
        time_in_force: TimeInForce | None = TimeInForce.GTC,
        tp_order_type: OrderType | None = None,
        tp_trigger_price: Decimal | None = None,
        tp_price: Decimal | None = None,
        tp_trigger_type: TriggerType | None = TriggerType.LAST_PRICE,
        sl_order_type: OrderType | None = None,
        sl_trigger_price: Decimal | None = None,
        sl_price: Decimal | None = None,
        sl_trigger_type: TriggerType | None = TriggerType.LAST_PRICE,
        **kwargs,
    ) -> Order:
        """Create a take profit and stop loss order"""
        raise NotImplementedError

    async def create_order(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal,
        time_in_force: TimeInForce = TimeInForce.GTC,
        reduce_only: bool = False,
        **kwargs,
    ) -> Order:
        """Create an order"""
        self._registry.register_tmp_order(
            order=Order(
                oid=oid,
                exchange=self._exchange_id,
                symbol=symbol,
                status=OrderStatus.INITIALIZED,
                amount=amount,
                type=type,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                timestamp=self._clock.timestamp_ms(),
                reduce_only=reduce_only,
            )
        )

        market = self._market.get(symbol)
        if not market:
            raise ValueError(
                f"Market {symbol} not found in exchange {self._exchange_id}"
            )

        if type.is_limit:
            time_in_force = HyperLiquidEnumParser.to_hyperliquid_time_in_force(
                time_in_force
            )
        elif type.is_market:
            time_in_force = HyperLiquidTimeInForce.IOC
            bookl1 = self._cache.bookl1(symbol)
            if not bookl1:
                raise ValueError(
                    "Please subscribe to bookl1 first, Market requires bookl1"
                )
            if side.is_buy:
                price = self._price_to_precision(
                    symbol, bookl1.ask * (1 + self._max_slippage), "ceil"
                )
            else:
                price = self._price_to_precision(
                    symbol, bookl1.bid * (1 - self._max_slippage), "floor"
                )

        elif type.is_post_only:
            time_in_force = HyperLiquidTimeInForce.ALO

        params: HyperLiquidOrderRequest = {
            "a": int(market.baseId),
            "b": side.is_buy,
            "p": str(price),
            "s": str(amount),
            "r": reduce_only,
            "t": {"limit": {"tif": time_in_force.value}},
            "c": oid,  # client order id
        }
        params.update(kwargs)

        try:
            res = await self._api_client.place_orders(orders=[params])
            status = res.response.data.statuses[0]

            if status.error:
                error_msg = status.error
                self._log.error(
                    f"Failed to place order for {symbol}: {error_msg} params: {str(params)}"
                )
                order = Order(
                    exchange=self._exchange_id,
                    oid=oid,
                    timestamp=self._clock.timestamp_ms(),
                    symbol=symbol,
                    type=type,
                    side=side,
                    amount=amount,
                    price=float(price),
                    time_in_force=time_in_force,
                    status=OrderStatus.FAILED,
                    filled=Decimal(0),
                    remaining=amount,
                    reduce_only=reduce_only,
                )
            else:
                order_status = status.resting or status.filled
                order = Order(
                    oid=oid,
                    eid=str(order_status.oid),
                    exchange=self._exchange_id,
                    timestamp=self._clock.timestamp_ms(),
                    symbol=symbol,
                    type=type,
                    side=side,
                    amount=amount,
                    price=float(price),
                    time_in_force=time_in_force,
                    status=OrderStatus.PENDING,
                    filled=Decimal(0),
                    remaining=amount,
                    reduce_only=reduce_only,
                )
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(
                f"Failed to place order for {symbol}: {error_msg} params: {str(params)}"
            )
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=symbol,
                type=type,
                side=side,
                amount=amount,
                price=float(price),
                time_in_force=time_in_force,
                status=OrderStatus.FAILED,
                filled=Decimal(0),
                remaining=amount,
                reduce_only=reduce_only,
            )
        self.order_status_update(order)

    async def create_batch_orders(
        self,
        orders: List[BatchOrderSubmit],
    ) -> List[Order]:
        """Create a batch of orders"""

        batch_orders: List[HyperLiquidOrderRequest] = []
        for order in orders:
            self._registry.register_tmp_order(
                order=Order(
                    oid=order.oid,
                    exchange=self._exchange_id,
                    symbol=order.symbol,
                    status=OrderStatus.INITIALIZED,
                    amount=order.amount,
                    type=order.type,
                    price=float(order.price) if order.price else None,
                    time_in_force=order.time_in_force,
                    timestamp=self._clock.timestamp_ms(),
                    reduce_only=order.reduce_only,
                )
            )

            market = self._market.get(order.symbol)
            if not market:
                raise ValueError(
                    f"Market {order.symbol} not found in exchange {self._exchange_id}"
                )
            price = order.price
            if order.type.is_limit:
                time_in_force = HyperLiquidEnumParser.to_hyperliquid_time_in_force(
                    order.time_in_force
                )
            elif order.type.is_market:
                time_in_force = HyperLiquidTimeInForce.IOC
                bookl1 = self._cache.bookl1(order.symbol)
                if not bookl1:
                    raise ValueError(
                        "Please subscribe to bookl1 first, Market requires bookl1"
                    )
                if order.side.is_buy:
                    price = self._price_to_precision(
                        order.symbol, bookl1.ask * (1 + self._max_slippage), "ceil"
                    )
                else:
                    price = self._price_to_precision(
                        order.symbol, bookl1.bid * (1 - self._max_slippage), "floor"
                    )

            elif order.type.is_post_only:
                time_in_force = HyperLiquidTimeInForce.ALO

            params: HyperLiquidOrderRequest = {
                "a": int(market.baseId),
                "b": order.side.is_buy,
                "p": str(price),
                "s": str(order.amount),
                "r": order.reduce_only,
                "t": {"limit": {"tif": time_in_force.value}},
                "c": order.oid,  # client order id
            }
            params.update(order.kwargs)
            batch_orders.append(params)

        try:
            res = await self._api_client.place_orders(orders=batch_orders)
            for order, status in zip(orders, res.response.data.statuses):
                if status.error:
                    error_msg = status.error
                    self._log.error(
                        f"Failed to place order for {order.symbol}: {error_msg} params: {str(params)}"
                    )
                    res_batch_order = Order(
                        oid=order.oid,
                        exchange=self._exchange_id,
                        timestamp=self._clock.timestamp_ms(),
                        symbol=order.symbol,
                        type=order.type,
                        side=order.side,
                        amount=order.amount,
                        price=float(order.price),
                        time_in_force=order.time_in_force,
                        status=OrderStatus.FAILED,
                        filled=Decimal(0),
                        remaining=order.amount,
                        reduce_only=order.reduce_only,
                    )
                else:
                    order_status = status.resting or status.filled
                    res_batch_order = Order(
                        exchange=self._exchange_id,
                        timestamp=self._clock.timestamp_ms(),
                        eid=str(order_status.oid),
                        oid=order.oid,
                        symbol=order.symbol,
                        type=order.type,
                        side=order.side,
                        amount=order.amount,
                        price=float(order.price),
                        time_in_force=order.time_in_force,
                        status=OrderStatus.PENDING,
                        filled=Decimal(0),
                        remaining=order.amount,
                        reduce_only=order.reduce_only,
                    )
                self.order_status_update(res_batch_order)

        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error creating batch orders: {error_msg}")
            for order in orders:
                order = Order(
                    exchange=self._exchange_id,
                    timestamp=self._clock.timestamp_ms(),
                    symbol=order.symbol,
                    oid=order.oid,
                    type=order.type,
                    side=order.side,
                    amount=order.amount,
                    price=float(order.price) if order.price else None,
                    time_in_force=order.time_in_force,
                    status=OrderStatus.FAILED,
                    filled=Decimal(0),
                    remaining=order.amount,
                    reduce_only=order.reduce_only,
                )
                self.order_status_update(order)

    async def cancel_order(self, oid: str, symbol: str, **kwargs) -> Order:
        """Cancel an order"""
        market = self._market.get(symbol)
        if not market:
            raise ValueError(
                f"Market {symbol} not found in exchange {self._exchange_id}"
            )

        params: HyperLiquidCloidCancelRequest = {
            "asset": int(market.baseId),
            "cloid": oid,
        }

        try:
            res = await self._api_client.cancel_orders_by_cloid(cancels=[params])
            status = res.response.data.statuses[0]
            if status == "success":
                order = Order(
                    oid=oid,
                    exchange=self._exchange_id,
                    timestamp=self._clock.timestamp_ms(),
                    symbol=symbol,
                    status=OrderStatus.CANCELING,
                )
            else:
                error_msg = status.error if status.error else "Unknown error"
                self._log.error(
                    f"Failed to cancel order for {symbol}: {error_msg} params: {str(params)}"
                )
                order = Order(
                    oid=oid,
                    exchange=self._exchange_id,
                    timestamp=self._clock.timestamp_ms(),
                    symbol=symbol,
                    status=OrderStatus.CANCEL_FAILED,
                )
            self.order_status_update(order)
        except Exception as e:
            error_msg = f"{e.__class__.__name__}: {str(e)}"
            self._log.error(f"Error canceling order: {error_msg} params: {str(params)}")
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=symbol,
                status=OrderStatus.CANCEL_FAILED,
            )
            self.order_status_update(order)

    async def modify_order(
        self,
        oid: str,
        symbol: str,
        order_id: str,
        side: OrderSide | None = None,
        price: Decimal | None = None,
        amount: Decimal | None = None,
        **kwargs,
    ) -> Order:
        """Modify an order"""
        raise NotImplementedError

    async def cancel_all_orders(self, symbol: str) -> bool:
        """Cancel all orders"""
        pass

    async def create_order_ws(
        self,
        oid: str,
        symbol: str,
        side: OrderSide,
        type: OrderType,
        amount: Decimal,
        price: Decimal,
        time_in_force: TimeInForce = TimeInForce.GTC,
        reduce_only: bool = False,
        **kwargs,
    ):
        """Create an order via WebSocket API"""
        self._registry.register_tmp_order(
            order=Order(
                oid=oid,
                exchange=self._exchange_id,
                symbol=symbol,
                status=OrderStatus.INITIALIZED,
                amount=amount,
                type=type,
                price=float(price) if price else None,
                time_in_force=time_in_force,
                timestamp=self._clock.timestamp_ms(),
                reduce_only=reduce_only,
            )
        )

        market = self._market.get(symbol)
        if not market:
            raise ValueError(
                f"Market {symbol} not found in exchange {self._exchange_id}"
            )

        if type.is_limit:
            time_in_force = HyperLiquidEnumParser.to_hyperliquid_time_in_force(
                time_in_force
            )
        elif type.is_market:
            time_in_force = HyperLiquidTimeInForce.IOC
            bookl1 = self._cache.bookl1(symbol)
            if not bookl1:
                raise ValueError(
                    "Please subscribe to bookl1 first, Market requires bookl1"
                )
            if side.is_buy:
                price = self._price_to_precision(
                    symbol, bookl1.ask * (1 + self._max_slippage), "ceil"
                )
            else:
                price = self._price_to_precision(
                    symbol, bookl1.bid * (1 - self._max_slippage), "floor"
                )
        elif type.is_post_only:
            time_in_force = HyperLiquidTimeInForce.ALO

        params: HyperLiquidOrderRequest = {
            "a": int(market.baseId),
            "b": side.is_buy,
            "p": str(price),
            "s": str(amount),
            "r": reduce_only,
            "t": {"limit": {"tif": time_in_force.value}},
            "c": oid,  # client order id
        }
        params.update(kwargs)

        await self._ws_api_client.place_order(id=oid, orders=[params])

    async def cancel_order_ws(self, oid: str, symbol: str, **kwargs):
        market = self._market.get(symbol)
        if not market:
            raise ValueError(
                f"Market {symbol} not found in exchange {self._exchange_id}"
            )

        params: HyperLiquidCloidCancelRequest = {
            "asset": int(market.baseId),
            "cloid": oid,
        }

        await self._ws_api_client.cancel_orders_by_cloid(id=oid, cancels=[params])

    def _ws_msg_handler(self, raw: bytes):
        """Handle WebSocket messages"""
        try:
            ws_msg: HyperLiquidWsMessageGeneral = self._ws_msg_general_decoder.decode(
                raw
            )
            # if ws_msg.channel == "pong":
            #     self._ws_client._transport.notify_user_specific_pong_received()
            #     self._log.debug("Pong received")
            #     return

            if ws_msg.channel == "orderUpdates":
                self._parse_order_update(raw)
            elif ws_msg.channel == "user":
                self._parse_user_events(raw)
        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding WebSocket message: {str(raw)} {e}")
            return

    def _ws_api_msg_handler(self, raw: bytes):
        """Handle WebSocket API messages for order operations"""
        try:
            ws_msg: HyperLiquidWsApiGeneralMsg = self._ws_api_msg_decoder.decode(raw)

            # if ws_msg.is_pong:
            #     self._ws_api_client._transport.notify_user_specific_pong_received()
            #     self._log.debug("API Pong received")
            #     return

            if ws_msg.is_order_response and ws_msg.data:
                self._parse_ws_api_response(ws_msg.data)

        except msgspec.DecodeError as e:
            self._log.error(f"Error decoding WebSocket API message: {str(raw)} {e}")
            return

    def _parse_ws_api_response(self, data: HyperLiquidWsApiMessageData):
        """Parse WebSocket API response and create Order objects"""
        request_id = data.id
        oid = self._ws_api_client._int_to_oid(request_id)
        response = data.response

        if response.payload.status != "ok":
            self._log.error(
                f"WebSocket API error for request {request_id}: {response.payload.status}"
            )
            return

        response_data = response.payload.response
        if response_data.type == "order":
            # Handle order placement response
            order_data: HyperLiquidWsApiOrderData = response_data.data
            for status in order_data.statuses:
                self._create_order_from_ws_response(oid, status)
        elif response_data.type == "cancel":
            # Handle order cancellation response
            cancel_data: HyperLiquidWsApiOrderData = response_data.data
            for status in cancel_data.statuses:
                self._create_cancel_order_from_ws_response(oid, status)

    def _create_order_from_ws_response(
        self, oid: str, status: HyperLiquidWsApiOrderStatus
    ) -> Order | None:
        """Create Order object from WebSocket API order response"""
        temp_order = self._registry.get_tmp_order(oid)
        if not temp_order:
            return

        if status.error:
            self._log.error(f"Order placement failed for {oid}: {status.error}")
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=temp_order.symbol,
                type=temp_order.type,
                side=temp_order.side,
                amount=temp_order.amount,
                price=temp_order.price,
                time_in_force=temp_order.time_in_force,
                status=OrderStatus.FAILED,
                filled=Decimal(0),
                remaining=temp_order.amount,
                reduce_only=temp_order.reduce_only,
            )
            self.order_status_update(order)
        else:
            eid = None
            if status.filled:
                eid = status.filled.oid
            elif status.resting:
                eid = status.resting.oid

            if eid:
                order = Order(
                    oid=oid,
                    exchange=self._exchange_id,
                    timestamp=self._clock.timestamp_ms(),
                    eid=str(eid),
                    symbol=temp_order.symbol,
                    type=temp_order.type,
                    side=temp_order.side,
                    amount=temp_order.amount,
                    price=temp_order.price,
                    time_in_force=temp_order.time_in_force,
                    status=OrderStatus.PENDING,
                    filled=Decimal(0),
                    remaining=temp_order.amount,
                    reduce_only=temp_order.reduce_only,
                )
                self.order_status_update(order)

    def _create_cancel_order_from_ws_response(
        self, oid: str, status: str | HyperLiquidWsApiOrderStatus
    ) -> Order | None:
        """Create Order object from WebSocket API cancel response"""
        temp_order = self._registry.get_tmp_order(oid)
        if not temp_order:
            return

        if isinstance(status, str) and status == "success":
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=temp_order.symbol,
                status=OrderStatus.CANCELING,
            )
        elif isinstance(status, HyperLiquidWsApiOrderStatus):
            self._log.error(f"Order cancellation failed for {oid}: {status.error}")
            order = Order(
                oid=oid,
                exchange=self._exchange_id,
                timestamp=self._clock.timestamp_ms(),
                symbol=temp_order.symbol,
                status=OrderStatus.CANCEL_FAILED,
            )
        self.order_status_update(order)

    def _parse_user_events(self, raw: bytes):
        user_fills_msg = self._ws_msg_user_events_decoder.decode(raw)
        self._log.debug(f"User fills received: {str(user_fills_msg)}")
        if not user_fills_msg.data.fills:
            return

        fills = user_fills_msg.data.fills
        for fill in fills:
            symbol = self._market_id.get(fill.coin, None)
            if not symbol:
                return

            market = self._market[symbol]
            if market.swap:
                sz = Decimal(fill.sz) if fill.side.is_buy else -Decimal(fill.sz)
                signed_amount = Decimal(fill.startPosition) + sz
                position = Position(
                    symbol=symbol,
                    exchange=self._exchange_id,
                    signed_amount=signed_amount,
                    entry_price=float(
                        fill.px
                    ),  # NOTE: HyperLiquid does not provide entry price in fills, using fill price instead,
                    # I know we can use current position to calculate the entry price, but it costs performance
                    side=PositionSide.LONG
                    if signed_amount > Decimal("0")
                    else PositionSide.SHORT,
                )
                self._cache._apply_position(position)
                self._log.debug(f"Position updated: {str(position)}")
            else:
                current_balance = self._cache.get_balance(
                    account_type=self._account_type
                )

                quote_total = current_balance.balance_total.get(
                    market.quote, Decimal("0")
                )  # zero if not found
                if fill.side.is_buy:
                    base_amount = (
                        Decimal(fill.sz)
                        + Decimal(fill.startPosition)
                        - Decimal(fill.fee)
                    )
                    quote_amount = quote_total - Decimal(fill.px) * Decimal(fill.sz)
                    balances = [
                        Balance(asset=market.baseName, free=base_amount),
                        Balance(asset=market.quote, free=quote_amount),
                    ]

                else:
                    base_amount = Decimal(fill.startPosition) - Decimal(fill.sz)
                    quote_amount = (
                        quote_total
                        + Decimal(fill.px) * Decimal(fill.sz)
                        - Decimal(fill.fee)
                    )

                    balances = [
                        Balance(asset=market.baseName, free=base_amount),
                        Balance(asset=market.quote, free=quote_amount),
                    ]

                self._log.debug(
                    f"\nbase: {market.baseName}, free: {base_amount}\n"
                    f"quote: {market.quote}, free: {quote_amount}"
                )

                self._cache._apply_balance(
                    account_type=self._account_type, balances=balances
                )

    def _parse_order_update(self, raw: bytes):
        order_msg: HyperLiquidWsOrderUpdatesMsg = (
            self._ws_msg_order_updates_decoder.decode(raw)
        )
        self._log.debug(f"Order update received: {str(order_msg)}")

        for data in order_msg.data:
            tmp_order = self._registry.get_tmp_order(data.order.cloid)
            if not tmp_order:
                continue

            id = data.order.coin
            symbol = self._market_id[id]
            order_status = HyperLiquidEnumParser.parse_order_status(data.status)
            order = Order(
                exchange=self._exchange_id,
                eid=str(data.order.oid),
                oid=data.order.cloid,
                symbol=symbol,
                type=tmp_order.type,  # HyperLiquid only have limit orders
                side=OrderSide.BUY if data.order.side.is_buy else OrderSide.SELL,
                amount=Decimal(data.order.origSz),
                price=float(data.order.limitPx),
                average=float(data.order.limitPx),
                status=order_status,
                filled=Decimal(data.order.origSz) - Decimal(data.order.sz),
                remaining=Decimal(data.order.sz),
                timestamp=data.order.timestamp,
                reduce_only=tmp_order.reduce_only,
            )
            self._log.debug(f"Parsed order: {str(order)}")
            self.order_status_update(order)
